"""Custom SpanSelector widget that has a higher rectangle without visible
up/down borders."""
# pylint: disable=invalid-name

import numpy as np

from matplotlib.widgets import AxesWidget, _SelectorWidget
from matplotlib.patches import Rectangle, Wedge
from matplotlib.transforms import blended_transform_factory

from gxps.utility import Observable


class PointSelector(_SelectorWidget):
    """Select a point on the canvas."""
    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-instance-attributes
    # pylint: disable=invalid-name
    # pylint: disable=attribute-defined-outside-init
    def __init__(self, ax, onselect, useblit=False, onmove_callback=None,
                 button=None):
        _SelectorWidget.__init__(
            self, ax, onselect, useblit=useblit, button=button)

        self.pressv = None

        self.onmove_callback = onmove_callback

        # Needed when dragging out of axes
        self.prev = (0, 0)

        # Reset canvas so that `new_axes` connects events.
        self.canvas = None
        self.new_axes(ax)

    def new_axes(self, ax):
        """Set SpanSelector to operate on a new Axes"""
        self.ax = ax
        if self.canvas is not ax.figure.canvas:
            if self.canvas is not None:
                self.disconnect_events()

            self.canvas = ax.figure.canvas
            self.connect_default_events()

    def ignore(self, event):
        """return *True* if *event* should be ignored"""
        return _SelectorWidget.ignore(self, event) or not self.visible

    def _press(self, event):
        """on button press event"""
        x0, y0 = self._get_data(event)
        self.pressv = (x0, y0)
        return False

    def _release(self, event):
        """on button release event"""
        if self.pressv is None:
            return True
        self.buttonDown = False

        x0, y0 = self.pressv

        x, y = self._get_data(event)

        self.onselect(x0, y0, x, y)
        self.pressv = None
        return False

    def _onmove(self, event):
        """on motion notify event"""
        if self.pressv is None:
            return True
        x, y = self._get_data(event)
        if x is None:
            return True
        x0, y0, = self.pressv

        if self.onmove_callback is not None:
            self.onmove_callback(x0, y0, x, y)

        self.update()
        return False


class PeakSelector(_SelectorWidget):
    """Draw a Peak as triangle."""
    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-instance-attributes
    # pylint: disable=invalid-name
    # pylint: disable=attribute-defined-outside-init
    def __init__(self, ax, onselect, minfwhm=None, minamp=None, useblit=False,
                 wedgeprops=None, onmove_callback=None, peak_stays=False,
                 button=None, limits=None):
        _SelectorWidget.__init__(
            self, ax, onselect, useblit=useblit, button=button)

        if minfwhm is not None or minamp is not None:
            raise NotImplementedError

        if wedgeprops is None:
            wedgeprops = dict(facecolor='red', alpha=0.5, fill=True)

        wedgeprops['animated'] = self.useblit

        self.wedge = None
        self.pressv = None

        if limits is None:
            limits = (-np.inf, np.inf)
        self.limits = limits

        self.wedgeprops = wedgeprops
        self.onmove_callback = onmove_callback
        self.minfwhm = minfwhm
        self.minamp = minamp
        self.peak_stays = peak_stays

        # Needed when dragging out of axes
        self.prev = (0, 0)

        # Reset canvas so that `new_axes` connects events.
        self.canvas = None
        self.new_axes(ax)

    def new_axes(self, ax):
        """Set SpanSelector to operate on a new Axes"""
        self.ax = ax
        if self.canvas is not ax.figure.canvas:
            if self.canvas is not None:
                self.disconnect_events()

            self.canvas = ax.figure.canvas
            self.connect_default_events()

        self.wedge = Wedge((0, 0), 1e10, 0, 0,
                           visible=False,
                           **self.wedgeprops)
        if self.peak_stays:
            self.stay_wedge = Wedge((0, 0), 1e10, 0, 0,
                                    visible=False,
                                    **self.wedgeprops)
            self.stay_wedge.set_animated(False)
            self.ax.add_patch(self.stay_wedge)

        self.ax.add_patch(self.wedge)
        self.artists = [self.wedge]

    def set_wedgeprops(self, wedgeprops):
        """Custom: set new rectprops."""
        self.wedgeprops = wedgeprops
        self.new_axes(self.ax)

    def set_limits(self, limits):
        """Sets new limits. Peak will only be drawn when press event occurs
        inside these x values."""
        self.limits = limits

    def ignore(self, event):
        """return *True* if *event* should be ignored"""
        return _SelectorWidget.ignore(self, event) or not self.visible

    def _press(self, event):
        """on button press event"""
        x0, y0 = self._get_data(event)
        if not self.limits[0] <= x0 <= self.limits[1]:
            return True
        self.wedge.set_visible(self.visible)
        if self.peak_stays:
            self.stay_wedge.set_visible(False)
            # really force a draw so that the stay rect is not in
            # the blit background
            if self.useblit:
                self.canvas.draw()
        self.pressv = (x0, y0)
        self.wedge.set_center((x0, y0))
        return False

    def _release(self, event):
        """on button release event"""
        if self.pressv is None:
            return True
        self.buttonDown = False

        self.wedge.set_visible(False)

        if self.peak_stays:
            self.stay_wedge.set_center(self.wedge.center)
            self.stay_wedge.set_radius(self.wedge.r)
            self.stay_wedge.set_theta1(self.wedge.theta1)
            self.stay_wedge.set_theta2(self.wedge.theta2)
            self.stay_wedge.set_visible(True)

        self.canvas.draw_idle()

        x0, y0, = self.pressv
        center = x0
        amplitude = y0

        x, y = self._get_data(event)
        angle = abs(np.arctan((x - x0) / (y - y0)))

        self.onselect(center, amplitude, angle)
        self.pressv = None
        return False

    def _onmove(self, event):
        """on motion notify event"""
        if self.pressv is None:
            return True
        x, y = self._get_data(event)
        if x is None:
            return True
        x0, y0, = self.pressv

        angle = abs(np.arctan((x - x0) / (y - y0)))
        self.wedge.set_theta1(np.rad2deg(-angle) - 90)
        self.wedge.set_theta2(np.rad2deg(angle) - 90)

        if self.onmove_callback is not None:
            center = x0
            amplitude = y0
            self.onmove_callback(center, amplitude, angle)

        self.update()
        return False


class SpanSelector(_SelectorWidget):
    """Custom SpanSelector."""
    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-arguments
    # pylint: disable=attribute-defined-outside-init
    # pylint: disable=invalid-name
    def __init__(self, ax, onselect, direction, minspan=None, useblit=False,
                 rectprops=None, onmove_callback=None, span_stays=False,
                 button=None):

        _SelectorWidget.__init__(self, ax, onselect, useblit=useblit,
                                 button=button)

        if rectprops is None:
            rectprops = dict(facecolor='red', alpha=0.5)

        rectprops['animated'] = self.useblit

        if direction not in ['horizontal', 'vertical']:
            msg = "direction must be in [ 'horizontal' | 'vertical' ]"
            raise ValueError(msg)
        self.direction = direction

        self.rect = None
        self.pressv = None

        self.rectprops = rectprops
        self.onmove_callback = onmove_callback
        self.minspan = minspan
        self.span_stays = span_stays

        # Needed when dragging out of axes
        self.prev = (0, 0)

        # Reset canvas so that `new_axes` connects events.
        self.canvas = None
        self.new_axes(ax)

    def new_axes(self, ax):
        """Set SpanSelector to operate on a new Axes"""
        self.ax = ax
        if self.canvas is not ax.figure.canvas:
            if self.canvas is not None:
                self.disconnect_events()

            self.canvas = ax.figure.canvas
            self.connect_default_events()

        if self.direction == 'horizontal':
            trans = blended_transform_factory(self.ax.transData,
                                              self.ax.transAxes)
            w, h = 0, 2
        else:
            trans = blended_transform_factory(self.ax.transAxes,
                                              self.ax.transData)
            w, h = 1, 0
        self.rect = Rectangle((0, -0.5), w, h,
                              transform=trans,
                              visible=False,
                              **self.rectprops)
        if self.span_stays:
            self.stay_rect = Rectangle((0, 0), w, h,
                                       transform=trans,
                                       visible=False,
                                       **self.rectprops)
            self.stay_rect.set_animated(False)
            self.ax.add_patch(self.stay_rect)

        self.ax.add_patch(self.rect)
        self.artists = [self.rect]

    def set_rectprops(self, rectprops):
        """Custom: set new rectprops."""
        self.rectprops = rectprops
        self.new_axes(self.ax)

    def ignore(self, event):
        """return *True* if *event* should be ignored"""
        return _SelectorWidget.ignore(self, event) or not self.visible

    def _press(self, event):
        """on button press event"""
        if self.ignore(event):
            return True
        self.rect.set_visible(self.visible)
        if self.span_stays:
            self.stay_rect.set_visible(False)
            # really force a draw so that the stay rect is not in
            # the blit background
            if self.useblit:
                self.canvas.draw()
        xdata, ydata = self._get_data(event)
        if self.direction == 'horizontal':
            self.pressv = xdata
        else:
            self.pressv = ydata
        return False

    def _release(self, event):
        """on button release event"""
        if self.ignore(event):
            return True
        if self.pressv is None:
            return True
        self.buttonDown = False

        self.rect.set_visible(False)

        if self.span_stays:
            self.stay_rect.set_x(self.rect.get_x())
            self.stay_rect.set_y(self.rect.get_y())
            self.stay_rect.set_width(self.rect.get_width())
            self.stay_rect.set_height(self.rect.get_height())
            self.stay_rect.set_visible(True)

        self.canvas.draw_idle()
        vmin = self.pressv
        xdata, ydata = self._get_data(event)
        if self.direction == 'horizontal':
            vmax = xdata or self.prev[0]
        else:
            vmax = ydata or self.prev[1]

        if vmin > vmax:
            vmin, vmax = vmax, vmin
        span = vmax - vmin
        if self.minspan is not None and span < self.minspan:
            return True
        self.onselect(vmin, vmax)
        self.pressv = None
        return False

    def _onmove(self, event):
        """on motion notify event"""
        if self.ignore(event):
            return True
        if self.pressv is None:
            return True
        x, y = self._get_data(event)
        if x is None:
            return True

        self.prev = x, y
        if self.direction == 'horizontal':
            v = x
        else:
            v = y

        minv, maxv = v, self.pressv
        if minv > maxv:
            minv, maxv = maxv, minv
        if self.direction == 'horizontal':
            self.rect.set_x(minv)
            self.rect.set_width(maxv - minv)
        else:
            self.rect.set_y(minv)
            self.rect.set_height(maxv - minv)

        if self.onmove_callback is not None:
            vmin = self.pressv
            xdata, ydata = self._get_data(event)
            if self.direction == 'horizontal':
                vmax = xdata or self.prev[0]
            else:
                vmax = ydata or self.prev[1]

            if vmin > vmax:
                vmin, vmax = vmax, vmin
            self.onmove_callback(vmin, vmax)

        self.update()
        return False


class DraggableVLine(Observable, AxesWidget):
    """A draggable vertical line in the plot."""
    _signals = ("changed-vline", )
    def __init__(self, line, spectrum):
        self.line = line
        super().__init__(line.axes)
        self.press = None
        self.background = None
        self.spectrum = spectrum

        self.connect()

    def connect(self):
        """Connect to the signals."""
        self.cidpress = self.canvas.mpl_connect(
            "button_press_event", self.on_press)
        self.cidrelease = self.canvas.mpl_connect(
            "button_release_event", self.on_release)
        self.cidmotion = self.canvas.mpl_connect(
            "motion_notify_event", self.on_motion)

    def disconnect(self):
        """Disconnects from canvas signals."""
        self.canvas.mpl_disconnect(self.cidpress)
        self.canvas.mpl_disconnect(self.cidrelease)
        self.canvas.mpl_disconnect(self.cidmotion)

    def on_press(self, event):
        """When the mouse button is pressed."""
        if not self.canvas.widgetlock.available(self):
            return
        if event.inaxes != self.line.axes:
            return
        if not self.line.contains(event)[0]:
            return

        self.press = self.line.get_xdata(), event.xdata, event.ydata

        self.line.set_animated(True)
        self.canvas.draw()
        self.background = self.canvas.copy_from_bbox(self.line.axes.bbox)
        self.line.axes.draw_artist(self.line)
        self.canvas.blit(self.line.axes.bbox)
        self.canvas.widgetlock(self)
        self.emit(
            "changed-vline",
            attr="press",
            value=self.line.get_xdata(),
            data=self.spectrum,
            noqueue=True
        )

    def on_release(self, _event):
        """When the mouse button is released."""
        if self.press is None:
            return
        if not self.canvas.widgetlock.available(self):
            return

        old_value = self.press[0]
        self.press = None

        self.line.set_animated(False)
        self.background = None
        self.canvas.widgetlock.release(self)
        self.emit(
            "changed-vline",
            attr="release",
            old_value=old_value,
            value=self.line.get_xdata(),
            data=self.spectrum,
            noqueue=True
        )

    def on_motion(self, event):
        """When the mouse is moved in pressed state."""
        if self.press is None:
            return False
        if not self.canvas.widgetlock.isowner(self):
            return False
        if event.inaxes != self.line.axes:
            return False

        xdata, xpress, _ = self.press
        self.line.set_xdata(xdata)
        xdiff = event.xdata - xpress
        self.line.set_xdata([self.line.get_xdata()[0] + xdiff] * 2)

        self.canvas.restore_region(self.background)
        self.line.axes.draw_artist(self.line)
        self.canvas.blit(self.line.axes.bbox)
        return True


# class DraggableAttributeLine(DraggableVLine):
#     """Takes a line marking a region boundary and makes it draggable."""
#     def __init__(self, line, ID, attr, callback):
#         super().__init__(line)
#         self.ID = ID
#         self.attr = attr
#         self.callback = callback
#
#     def on_release(self, event):
#         """When the mouse button is released."""
#         doit = super().on_release(event)
#         if doit:
#             cbargs = {self.attr: self.line.get_xdata()[0]}
#             self.callback(self.ID, **cbargs)
#             self.canvas.draw()



"""Provides functions for processing spectrum data."""
# pylint: disable=invalid-name
# pylint: disable=logging-format-interpolation

import logging

import numpy as np


LOG = logging.getLogger(__name__)


class IgnoreUnderflow:
    """Context manager for suppressing numpy underflow runtime errors."""
    # pylint: disable=too-few-public-methods
    def __init__(self):
        self.old_settings = {}

    def __enter__(self):
        self.old_settings = np.seterr(under="ignore")

    def __exit__(self, exc_type, exc_value, traceback):
        np.seterr(**self.old_settings)


def calculate_background(bg_type, bg_bounds, energy, intensity):
    """Calculates a numpy array representing the background."""
    background = intensity.copy()
    if bg_type == "none":
        return np.zeros(energy.shape)
    if bg_bounds.size == 0:
        bg_bounds = [energy.min(), energy.max()]
    for lower, upper in zip(bg_bounds[0::2], bg_bounds[1::2]):
        idx1, idx2 = sorted([
            np.searchsorted(energy, lower),
            np.searchsorted(energy, upper)
        ])
        if idx1 == idx2:
            continue
        if bg_type == "shirley":
            try:
                background[idx1:idx2] = shirley(
                    energy[idx1:idx2], intensity[idx1:idx2]
                )
            except FloatingPointError:
                LOG.warning("shirley: division by zero")
        elif bg_type == "linear":
            background[idx1:idx2] = linear_bg(
                energy[idx1:idx2], intensity[idx1:idx2]
            )
        elif bg_type == "tougaard":
            raise NotImplementedError("Tougaard not implemented")
        else:
            raise ValueError("Unknown background type '{}'".format(bg_type))
    return background


def intensity_at_energy(energy, intensity, energy_value):
    """Gives back the intensity value at energy_value."""
    idx = np.searchsorted(energy, energy_value)
    return intensity[idx]


def shirley(energy, intensity, tol=1e-5, maxit=20):
    """Calculates shirley background for given x, y values."""
    if energy[-1] < energy[0]:
        raise ValueError("Energy not increasing.")
    if not is_equidistant(energy):
        raise ValueError("Energy not evenly spaced.")

    energy = energy[::-1]
    intensity = intensity[::-1]
    np.seterr(all="raise")

    background = np.ones(energy.shape) * intensity[-1]
    integral = np.zeros(energy.shape)
    spacing = (energy[-1] - energy[0]) / (len(energy) - 1)

    rest = intensity - background
    ysum = rest.sum() - np.cumsum(rest)
    integral = spacing * (ysum - 0.5 * (rest + rest[-1]))

    for _ in range(maxit):
        rest = intensity - background
        integral = spacing * (rest.sum() - np.cumsum(rest))
        bnew = (
            (intensity[0] - intensity[-1])
            * integral / integral[0]
            + intensity[-1]
        )
        if np.linalg.norm((bnew - background) / intensity[0]) < tol:
            background = bnew
            break
        else:
            background = bnew
    else:
        LOG.warning("shirley: Max iterations exceeded before convergence.")

    return background[::-1]

def linear_bg(energy, intensity):
    """Calculates linear background for given x, y values."""
    samples = len(energy)
    background = np.linspace(intensity[0], intensity[-1], samples)
    return background

def calculate_normalization_divisor(norm_type, norm_div, _energy, intensity):
    """Calculates normalization divisor."""
    new_divisor = 1.0
    if norm_type == "highest":
        new_divisor = intensity.max()
    elif norm_type == "manual":
        new_divisor = norm_div
    elif norm_type == "none":
        pass
    elif norm_type == "high_energy":
        span = intensity.max() - intensity.min()
        for i, intensity_value in enumerate(intensity[::-1]):
            if i > 0 and abs(intensity_value - intensity[-1]) > 0.05 * span:
                new_divisor = intensity[-i:].mean()
                break
        else:
            new_divisor = intensity[-1]
    elif norm_type == "low_energy":
        span = intensity.max() - intensity.min()
        for i, intensity_value in enumerate(intensity):
            if i > 0 and abs(intensity_value - intensity[0]) > 0.05 * span:
                new_divisor = intensity[:i].mean()
                break
        else:
            new_divisor = intensity[0]
    else:
        ValueError("Invalid normalization type '{}'".format(norm_type))
    return new_divisor

def is_equidistant(energy, tol=1e-08):
    """Returns True only when energy is equidistant."""
    spacings = np.unique(np.diff(energy))
    for spacing in spacings[1:]:
        if not np.isclose(spacing, spacings[0], atol=tol):
            return False
    return True

def make_increasing(energy, intensity):
    """Makes energy increasing and sorts intensity accordingly."""
    idxes = energy.argsort()
    incr_energy = energy[idxes]
    incr_intensity = intensity[idxes]
    return incr_energy, incr_intensity

def make_equidistant(energy, intensity):
    """Makes x, y pair so that x is equidistant."""
    spacings = np.unique(np.diff(energy))
    # try to eliminate spacings that are too small
    while np.isclose(0, spacings.min()):
        spacings.remove(spacings.min())
    if all([np.isclose(spacing, spacings[0]) for spacing in spacings]):
        return energy, intensity
    samples = int((energy.max() - energy.min()) / spacings.min())
    spaced_energy = np.linspace(energy.min(), energy.max(), samples)
    spaced_intensity = np.interp(spaced_energy, energy, intensity)
    return spaced_energy, spaced_intensity



"""Gtk.Application class."""
# pylint: disable=wrong-import-position
# pylint: disable=logging-format-interpolation

import logging

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, Gio, GLib

from gxps import __appname__, __version__
from gxps.xdg import DATA_DIR
from gxps.config import CONFIG, COLORS
from gxps.utility import EventBus
from gxps.spectrum import SpectrumContainer
from gxps.state import State
from gxps.control import CommandSender
from gxps.view import ViewManager

import gxps.widgets         # pylint: disable=unused-import


LOG = logging.getLogger(__name__)


class GXPS(Gtk.Application):
    """Application class organising user interaction."""
    # pylint: disable=arguments-differ
    # pylint: disable=too-many-instance-attributes
    def __init__(self):
        app_id = "com.github.schachmett.{}".format(__appname__.lower())
        super().__init__(
            application_id=app_id,
            flags=Gio.ApplicationFlags.HANDLES_COMMAND_LINE
        )
        GLib.set_application_name(__appname__)
        self.set_resource_base_path(str(DATA_DIR / "assets"))

        self.builder = Gtk.Builder.new()
        self.builder.add_from_file(str(DATA_DIR / "ui/gxps.glade"))
        self.builder.add_from_file(str(DATA_DIR / "ui/menus.ui"))
        self.get_widget = self.builder.get_object

        self.win = None
        self.bus = None
        self.commandsender = None
        self.data = None
        self.state = None
        self._view = None

        self.add_main_option_entries([
            make_option("--verbosity", "-v", arg=GLib.OptionArg.INT,
                        description="Value from 1 (only errors) to 4 (debug)"),
            make_option("--version", description="show program version"),
            make_option("--clean", description="don't load previous file")
        ])

    def do_activate(self):
        """Instantiate MainWindow."""
        LOG.info("GXPS has version '{}'".format(__version__))
        # load the last used project file
        fname = CONFIG["IO"]["current-project"]
        self.commandsender(fname, "startup")

    def do_startup(self):
        """Adds actions."""
        LOG.info("Starting application...")
        Gtk.Application.do_startup(self)
        self.set_menubar(self.builder.get_object("menubar"))
        self.win = self.get_widget("main_window")
        self.win.startup(app=self)

        # initialize all the state objects and workers
        complement_builder(self.builder)
        self.bus = EventBus(default_policy="fire")
        self.data = SpectrumContainer()
        self.data.register_queue(self.bus)
        self.state = State(self, self.data)
        self.state.register_queue(self.bus)
        self._view = ViewManager(self, self.state, self.data)
        self.commandsender = CommandSender(
            self.bus,
            self.data,
            self.state,
            self.get_widget
        )

        callbacks = [
            "on_spectrum_view_search_entry_changed",
            "on_spectrum_view_search_combo_changed",
            "on_spectrum_view_button_press_event",
            "on_spectrum_view_row_activated",
            "on_calibration_spinbutton_value_changed",
            "on_normalization_combo_changed",
            "on_normalization_entry_activate",
            "on_region_background_type_combo_changed",
            "on_peak_entry_activate",
            "on_peak_name_entry_changed",
            "on_peak_view_row_activated",
            "on_peak_model_combo_changed",
            "on_img_export_change"
        ]
        actions = [
            "project-new", "save-project", "save-project-as", "open-project",
            "merge-project", "import-spectra", "export-txt", "export-image",

            "edit-spectra", "remove-spectra", "avg-selected-spectra",

            "add-region", "remove-region", "clear-regions", "add-peak",
            "add-guessed-peak", "remove-peak", "clear-peaks", "fit",

            "show-selected-spectra", "show-atomlib", "center-plot", "pan-plot",
            "zoom-plot",

            "view-logfile", "edit-colors", "about"
        ]
        handlers = dict((key, (self.commandsender, key)) for key in callbacks)
        handlers["on_main_window_delete_event"] = self.on_quit
        self.builder.connect_signals(handlers)
        for name in actions:
            simple = Gio.SimpleAction.new(name, None)
            simple.connect("activate", self.commandsender, name)
            self.add_action(simple)
        simple = Gio.SimpleAction.new("quit", None)
        simple.connect("activate", self.on_quit)
        self.add_action(simple)

    def do_command_line(self, command_line):
        """Handles command line arguments"""
        Gtk.Application.do_command_line(self, command_line)
        options = command_line.get_options_dict()
        if options.contains("verbosity"):
            verb = options.lookup_value("verbosity", GLib.VariantType("i"))
            levels = (
                None,
                logging.ERROR,
                logging.WARNING,
                logging.INFO,
                logging.DEBUG
                )
            logging.getLogger().handlers[0].setLevel(levels[verb.unpack()])
        if options.contains("version"):
            print("{} version: {}".format(__appname__, __version__))
            self.quit()
            return 0
        if options.contains("clean"):
            CONFIG["IO"]["current-project"] = ""
        self.do_activate()
        return 0

    def on_quit(self, *_args):
        """Clean up, write configs, ask if user wants to save, and die."""
        if self.state.project_isaltered:
            self.commandsender("ask-for-save")
        if self.state.project_isaltered:
            return True
        xsize, ysize = self.win.get_size()
        xpos, ypos = self.win.get_position()
        CONFIG["Window"]["xsize"] = str(xsize)
        CONFIG["Window"]["ysize"] = str(ysize)
        CONFIG["Window"]["xpos"] = str(xpos)
        CONFIG["Window"]["ypos"] = str(ypos)
        CONFIG.save()
        COLORS.save()
        LOG.info("quitting...")
        self.quit()
        return False


def complement_builder(builder):
    """Do trivial GTK stuff that the builder can not do."""
    # Save confirmation dialog
    save_conf_dialog = builder.get_object("save_confirmation_dialog")
    save_conf_dialog.set_accels()
    # Plot navigation toolbar
    navbar = builder.get_object("plot_toolbar")
    navbar.startup(builder)
    # RSF dialog
    rsf_dialog = builder.get_object("rsf_dialog")
    rsf_entry = builder.get_object("rsf_entry")
    def apply_rsf(*_args):
        """Let the dialog send APPLY response."""
        rsf_dialog.emit("response", Gtk.ResponseType.APPLY)
    rsf_entry.connect("activate", apply_rsf)

def make_option(long_name, short_name=None, arg=GLib.OptionArg.NONE, **kwargs):
    """Make GLib option for the command line. Uses kwargs description, flags,
    arg_data and arg_description."""
    # surely something like this should exist inside PyGObject itself?!
    option = GLib.OptionEntry()
    option.long_name = long_name.lstrip('-')
    option.short_name = 0 if not short_name else ord(short_name.lstrip('-'))
    option.arg = arg
    option.flags = kwargs.get("flags", 0)
    option.arg_data = kwargs.get("arg_data", None)
    option.description = kwargs.get("description", None)
    option.arg_description = kwargs.get("arg_description", None)
    return option



"""Provides platform-independent access to directories and configuration."""
# pylint has a bug with pathlib: https://github.com/PyCQA/pylint/issues/1660
# pylint: disable=no-member

import os
from pathlib import Path
import sys
from gi.repository import GLib


GXPS_DIR = Path(os.path.realpath(__file__)).parents[1]
# The following hack works only if the gxps.py setting the env var is
# one directory above the gxps package
if os.getenv("GXPS_DIR") and sys.platform == "win32":
    GXPS_DIR = Path(os.getenv("GXPS_DIR"))
    print("INFO    : Source dir taken from environment variable.")
HOME_DIR = os.path.expanduser("~")

# detect if running without being installed
LOCAL_HACK = (GXPS_DIR / "data").is_dir() and sys.platform != "win32"
# detect if running from a virtualenv
VENV = sys.prefix != sys.base_prefix

DATA_DIR = Path(GLib.get_user_data_dir()) / "gxps"
if LOCAL_HACK or sys.platform == "win32":
    DATA_DIR = GXPS_DIR / "data"
if VENV:
    DATA_DIR = Path(sys.prefix) / "share/gxps"
DATA_DIRS = [DATA_DIR]
if os.getenv("XDG_DATA_DIRS"):
    XDG_DATA_DIRS = os.getenv("XDG_DATA_DIRS").split(os.pathsep)
    DATA_DIRS.extend([Path(d) / "gxps" for d in XDG_DATA_DIRS])

CONF_DIR = Path(GLib.get_user_config_dir()) / "gxps"
CONF_DIRS = [CONF_DIR]
if os.getenv("XDG_CONFIG_DIRS"):
    XDG_CONFIG_DIRS = os.getenv("XDG_CONFIG_DIRS").split(os.pathsep)
    CONF_DIRS.extend([Path(d) / "gxps" for d in XDG_CONFIG_DIRS])
if LOCAL_HACK or sys.platform == "win32":
    CONF_DIRS.append(GXPS_DIR / "data/config")

CACHE_DIR = Path(GLib.get_user_cache_dir()) / "gxps"

LOG_DIR = CACHE_DIR / "logs"
if sys.platform == "win32":
    LOG_DIR = CONF_DIR / "logs"
LOG_FILE = LOG_DIR / "gxps.log"


def _make_missing_dirs():
    if not DATA_DIR.is_dir():
        os.makedirs(str(DATA_DIR))
    if not CONF_DIR.is_dir():
        os.makedirs(str(CONF_DIR))
    if not CACHE_DIR.is_dir():
        os.makedirs(str(CACHE_DIR))
    if not LOG_DIR.is_dir():
        os.makedirs(str(LOG_DIR))



"""Manages database file and has import filters."""
# pylint: disable=logging-format-interpolation
# pylint: disable=global-statement

import re
import logging
import pickle
import sqlite3

import numpy as np

from gxps import __version__
from gxps.xdg import DATA_DIR


LOG = logging.getLogger(__name__)


def load_project(fname):
    """Loads project file."""
    with open(fname, "rb") as pfile:
        state = pickle.load(pfile)
    state = convert_older_version(state)
    return state

def convert_older_version(state):
    """If the project file has an older version, this should convert it.
    """
    if state[2] != __version__:
        LOG.warning("Project file has older version '{}'".format(state[2]))
    state.remove(state[2])
    return state

def save_project(fname, spectrum_container, gui_state):
    """Saves a StatefulSpectrumContainer as a file."""
    active_spectrum_idxs = []
    for spectrum in gui_state.active_spectra:
        s_idx = spectrum_container.spectra.index(spectrum)
        active_spectrum_idxs.append(s_idx)
    container = spectrum_container.deepcopy()
    spectra = container.spectra
    state = [
        spectra,
        active_spectrum_idxs,
        __version__
    ]
    with open(fname, "wb") as pfile:
        pickle.dump(state, pfile, pickle.HIGHEST_PROTOCOL)


def parse_spectrum_file(fname):
    """Checks file extension and calls appropriate parsing method."""
    specdicts = []
    with open(fname, "r") as sfile:
        firstline = sfile.readline()
    if fname.split(".")[-1] == "txt":
        if "Region" in firstline:
            for specdict in parse_eistxt(fname):
                specdicts.append(specdict)
        elif re.fullmatch(r"\d+\.\d+,\d+\n", firstline):
            specdicts.append(parse_simple_xy(fname, delimiter=","))
    elif fname.split(".")[-1] == "xy":
        if re.fullmatch(r"\d+\.\d+,\d+\n", firstline):
            delimiter = ","
        else:
            delimiter = None
        specdicts.append(parse_simple_xy(fname, delimiter=delimiter))
    if not specdicts:
        raise ValueError("Could not parse file '{}'".format(fname))
    return specdicts

def parse_simple_xy(fname, delimiter=None):
    """
    Parses the most simple x, y file with no header.
    """
    energy, intensity = np.genfromtxt(
        fname,
        delimiter=delimiter,
        unpack=True
    )
    specdict = {
        "filename": fname,
        "energy": energy,
        "intensity": intensity,
        "name": "S XY",
        "notes": "file {}".format(fname.split("/")[-1])
    }
    return specdict

def parse_eistxt(fname):
    """Splits Omicron EIS txt file."""
    splitregex = re.compile(r"^Region.*")
    skip_once_regex = re.compile(r"Layer.*")
    skip_regex = re.compile(r"^[0-9]+\s*False.*")
    split_eislines = []
    with open(fname, "br") as eisfile:
        for line in eisfile:
            line = line.decode("utf-8", "backslashreplace")
            if re.match(splitregex, line):
                split_eislines.append([])
                do_skip = False
            elif re.match(skip_regex, line):
                do_skip = True
            elif re.match(skip_once_regex, line):
                continue
            if not do_skip:
                split_eislines[-1].append(line)

    for data in split_eislines:
        energy, intensity = np.genfromtxt(
            data,
            skip_header=4,
            unpack=True
        )
        header = [line.split("\t") for line in data[:4]]
        specdict = {
            "filename": fname,
            "energy": energy,
            "intensity": intensity,
            "eis_region": int(header[1][0]),
            "name": "S {}".format(header[1][0]),
            "sweeps": int(header[1][6]),
            "dwelltime": float(header[1][7]),
            "pass_energy": float(header[1][9]),
            "notes": header[1][12],
        }
        yield specdict


def get_element_rsfs(element, source):
    """Return dictionary containing rsfs for a specific element / source.
    """
    source_photons = {
        "Al": 1486.3,
        "Mg": 1253.4
    }
    photon_energy = source_photons.get(source, None)
    if photon_energy is None:
        photon_energy = float(source)
    dbfname = str(DATA_DIR / "assets/rsf.db")
    with sqlite3.connect(dbfname) as database:
        cursor = database.cursor()
        sql = """
            SELECT IsAuger, Orbital, BE, RSF
            FROM Peak
            WHERE Element=? AND (Source=? OR Source="Any")
        """
        cursor.execute(sql, (element.title(), source))
        rsf_data = cursor.fetchall()
        rsf_dicts = []
        for isauger, orbital, energy, rsf in rsf_data:
            if isauger == 1.0:
                binding_energy = photon_energy - energy
                orbital = orbital.upper()
            else:
                binding_energy = energy
            rsf_dicts.append({
                "Element": element.title(),
                "Orbital": orbital,
                "BE": binding_energy,
                "RSF": rsf
            })
    return rsf_dicts


def export_txt(fname, spectrum):
    """Export given spectra and everything that belongs to it as txt."""
    column_stack = [
        spectrum.energy,
        spectrum.intensity,
    ]
    name = re.sub(r"\s+", "_", spectrum.name)
    header = "{:_<15}_Energy\t{:_<14}_intensity\t".format(name, name)
    if spectrum.background.any():
        column_stack.append(spectrum.background)
        header += "{:_<13}_background\t".format(name)
    if spectrum.fit.any():
        column_stack.append(spectrum.fit)
        header += "{:_<20}_fit\t".format(name)
    for peak in spectrum.peaks:
        column_stack.append(peak.intensity)
        peak_name = re.sub(r"\s+", "_", peak.name)
        header += "{:_<12}_{:_<3}_peakint\t".format(name, peak_name)
    data = np.column_stack(column_stack)
    np.savetxt(fname, data, delimiter="\t", header=header)



"""Class holding all the actions that can be invoked by the user.
"""
# pylint: disable=wrong-import-position
# pylint: disable=logging-format-interpolation
# pylint: disable=too-few-public-methods

import logging
import os
import sys
import re

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, Gdk

from gxps import __version__, __authors__, __website__, __appname__
from gxps.config import CONFIG, COLORS
from gxps.xdg import LOG_FILE, CONF_DIR
import gxps.io


LOG = logging.getLogger(__name__)


class Operator:
    """Meta class for objects that contain the main functions of the
    application as methods.
    """
    def __init__(self, get_widget, state, data, bus):
        self.get_widget = get_widget
        self.state = state
        self.data = data
        self.bus = bus


class Help(Operator):
    """Cares about extra windows etc."""
    @staticmethod
    def on_view_logfile(_action, *_args):
        """Views logfile in external text editor."""
        if sys.platform.startswith("linux"):
            os.system("xdg-open {}".format(str(LOG_FILE)))
        else:
            LOG.warning("logfile viewing only implemented for linux")

    @staticmethod
    def on_edit_colors(_action, *_args):
        """Views colors.ini file in external text editor."""
        if sys.platform.startswith("linux"):
            os.system("xdg-open {}".format(str(CONF_DIR / "colors.ini")))
        else:
            LOG.warning("color file editing only implemented for linux")

    def on_about(self, _widget, *_ignore):
        """Show 'About' dialog."""
        dialog = self.get_widget("about_dialog")
        dialog.run()
        LOG.debug("Showing 'About' window")
        dialog.hide()


class File(Operator):
    """Manages project files."""
    def startup(self, fname, *_args):
        """Opens a file or makes a new project if that file does not exist."""
        if fname:
            try:
                self.open(fname)
            except FileNotFoundError:
                LOG.warning("File '{}' not found".format(fname))
                self.new()
        else:
            self.new()
        self.bus.fire()

    def new(self):
        """Make new project."""
        self.data.clear()
        self.state.current_project = ""
        LOG.info("New project")

    def on_new(self, *_args):
        """User callback for making a new project."""
        if self.state.project_isaltered:
            self.ask_for_save()
        if self.state.project_isaltered:
            return
        self.new()
        self.bus.fire()

    def open(self, fname, merge=False):
        """Load project file."""
        spectra, active_idxs = gxps.io.load_project(fname)
        if not merge:
            self.data.clear()
        for spectrum in spectra:
            spectrum = self.data.add_spectrum(spectrum)
            spectrum.register_queue(self.bus)
            for peak in spectrum.peaks:
                peak.register_queue(self.bus)
        if not merge:
            self.state.active_spectra = [spectra[idx] for idx in active_idxs]
            self.state.current_project = fname
        LOG.info("Opened project file '{}'".format(fname))

    def on_open(self, *_args):
        """Let the user choose a project file to open and open it through
        self.open_project."""
        if self.state.project_isaltered:
            self.ask_for_save()
        if self.state.project_isaltered:
            return
        dialog = self.get_widget("open_project_dialog")
        project_dir = os.path.expandvars(CONFIG["IO"]["project-dir"])
        dialog.set_current_folder(project_dir)
        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            fname = dialog.get_filename()
            CONFIG["IO"]["project-dir"] = dialog.get_current_folder()
            self.open(fname)
        dialog.hide()
        self.bus.fire()

    def on_merge(self, *_args):
        """Merges a project file into the current project."""
        dialog = self.get_widget("merge_project_dialog")
        project_dir = os.path.expandvars(CONFIG["IO"]["project-dir"])
        dialog.set_current_folder(project_dir)
        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            fname = dialog.get_filename()
            CONFIG["IO"]["project-dir"] = dialog.get_current_folder()
            self.open(fname, merge=True)
        dialog.hide()
        self.bus.fire()

    def on_import(self, *_args):
        """Imports spectra from a data file."""
        dialog = self.get_widget("import_dialog")
        data_dir = os.path.expandvars(CONFIG["IO"]["data-dir"])
        dialog.set_current_folder(data_dir)
        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            CONFIG["IO"]["data-dir"] = dialog.get_current_folder()
            specdicts = []
            for fname in dialog.get_filenames():
                specdicts.extend(gxps.io.parse_spectrum_file(fname))
            for specdict in specdicts:
                spectrum = self.data.add_spectrum(**specdict)
                spectrum.register_queue(self.bus)
        dialog.hide()
        self.bus.fire()

    def ask_for_save(self):
        """Opens a AskForSaveDialog and then either saves the file or,
        if the user does not want to save, sets the project_isaltered
        to False. If the dialog is canceled, nothing happens and
        project_isaltered stays True."""
        dialog = self.get_widget("save_confirmation_dialog")
        response = dialog.run()
        if response == Gtk.ResponseType.YES:
            dialog.hide()
            self.on_save()
        elif response == Gtk.ResponseType.NO:
            dialog.hide()
            self.state.project_isaltered = False
        else:
            dialog.hide()
        self.bus.fire()

    def save(self, fname):
        """Saves project file."""
        gxps.io.save_project(fname, self.data, self.state)
        self.state.current_project = fname
        LOG.info("Saved project to '{}'".format(fname))

    def on_save(self, *_args):
        """Saves the current project."""
        project = self.state.current_project
        if not project or project == "Untitled":
            self.on_save_as()
        else:
            self.save(self.state.current_project)
        self.bus.fire()

    def on_save_as(self, *_args):
        """Saves the current project as a new file."""
        dialog = self.get_widget("save_project_dialog")
        project_dir = os.path.expandvars(CONFIG["IO"]["project-dir"])
        dialog.set_current_folder(project_dir)
        dialog.set_current_name("untitled.gxps")
        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            fname = dialog.get_filename()
            CONFIG["IO"]["project-dir"] = dialog.get_current_folder()
            self.save(fname)
        dialog.hide()
        self.bus.fire()

    def on_export_txt(self, *_args):
        """Exports the currently selected spectra and their fits to an ASCII
        file.
        """
        spectra = self.state.active_spectra
        dialog = self.get_widget("export_txt_dialog")
        project_dir = os.path.expandvars(CONFIG["IO"]["project-dir"])
        dialog.set_current_folder(project_dir)
        for spectrum in spectra:
            name = re.sub(r"\s+", "_", spectrum.name)
            dialog.set_current_name(name + ".txt")
            response = dialog.run()
            if response == Gtk.ResponseType.OK:
                fname = dialog.get_filename()
                gxps.io.export_txt(fname, spectrum)
        CONFIG["IO"]["project-dir"] = dialog.get_current_folder()
        dialog.hide()
        self.bus.fire()

    def on_export_image(self, *_args):
        """Exports the image currently displayed on the exporting canvas.
        """
        spectra = self.state.active_spectra
        canvas = self.get_widget("export_canvas")
        dialog = self.get_widget("export_img_dialog")
        project_dir = os.path.expandvars(CONFIG["IO"]["project-dir"])
        dialog.set_current_folder(project_dir)
        name = "_".join([re.sub(r"\s+", "-", s.name) for s in spectra])
        dialog.set_current_name(name + ".png")
        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            fname = dialog.get_filename()
            canvas.saveas(fname)
            CONFIG["IO"]["project-dir"] = dialog.get_current_folder()
            dialog.hide()
            return True
        dialog.hide()
        return False

    def on_start_image_exporter(self, *_args):
        """Exports the currently selected spectra and their fits as an image.
        """
        spectra = self.state.active_spectra
        dialog = self.get_widget("export_canvas_dialog")
        canvas = self.get_widget("export_canvas")
        canvas.plot_spectra(spectra)
        success = False
        while not success:
            response = dialog.run()
            if response == Gtk.ResponseType.OK:
                success = self.on_export_image()
            else:
                success = True
        dialog.hide()

    def on_change_image_exporter(self, *_args):
        """Changes the export canvas according to user settings."""
        canvas = self.get_widget("export_canvas")
        title = self.get_widget("img_export_title")
        xlabel = self.get_widget("img_export_xlabel")
        ylabel = self.get_widget("img_export_ylabel")
        canvas.ax.set_title(title.get_text())
        canvas.ax.set_xlabel(xlabel.get_text())
        canvas.ax.set_ylabel(ylabel.get_text())
        canvas.draw_idle()



class Edit(Operator):
    """Contains methods for user initiated data manipulation."""
    def on_remove_selected_spectra(self, *_args):
        """Removes selected spectra."""
        spectra = self.state.selected_spectra
        for spectrum in spectra:
            self.data.remove_spectrum(spectrum)
        self.bus.fire()

    def on_edit_spectra(self, *_args):
        """Opens an 'Edit' dialog for selected spectra."""
        spectra = self.state.selected_spectra
        if not spectra:
            return
        self.state.editing_spectra = spectra
        dialog = self.get_widget("edit_spectrum_dialog")
        response = dialog.run()
        if response == Gtk.ResponseType.APPLY:
            values = dialog.get_values()
            for spectrum in spectra:
                for attr, value in values.items():
                    spectrum.set_meta(attr, value)
        dialog.hide()
        self.state.editing_spectra = []
        self.bus.fire()

    def on_calibrate(self, *_args):
        """Changes the calibration for selected spectra."""
        adjustment = self.get_widget("calibration_spinbutton_adjustment")
        calibration = float(adjustment.get_value())
        for spectrum in self.state.active_spectra:
            spectrum.energy_calibration = calibration
        self.bus.fire()

    def on_normalize(self, *_args):
        """Changes the normalization for selected spectra."""
        combo = self.get_widget("normalization_combo")
        normid = combo.get_active_id()
        if normid is None:
            return
        norm_type = self.state.titles["norm_type_ids"].inverse[normid]
        for spectrum in self.state.active_spectra:
            spectrum.normalization_type = norm_type
        self.bus.fire()

    def on_normalize_manual(self, *_args):
        """Changes the normalization divisor directly."""
        entry = self.get_widget("normalization_entry")
        for spectrum in self.state.active_spectra:
            if spectrum.normalization_type != "manual":
                raise ValueError("Normalization is not set to manual")
            spectrum.normalization_divisor = 1 / float(entry.get_text())
        self.bus.fire()

    def on_change_bg(self, *_args):
        """Changes the background type."""
        active_spectra = self.state.active_spectra
        if len(active_spectra) < 1:
            return
        combo = self.get_widget("region_background_type_combo")
        bgid = combo.get_active_id()
        if bgid is None:
            return
        bg_type = self.state.titles["background_type_ids"].inverse[bgid]
        for spectrum in active_spectra:
            spectrum.background_type = bg_type
        self.bus.fire()


class Fit(Operator):
    """Methods for drawing peaks, fitting and peak and background
    manipulation."""
    def on_add_region(self, *_args):
        """Add two region boundaries to each of the active spectra."""
        navbar = self.get_widget("plot_toolbar")
        def add_region(emin, emax):
            """Add region"""
            for spectrum in self.state.active_spectra:
                spectrum.background_type = "shirley"
                spectrum.add_background_bounds(emin, emax)
            self.bus.fire()
        spanprops = {"edgecolor": COLORS["Plotting"]["region-vlines"], "lw": 2}
        navbar.get_span(add_region, **spanprops)

    def on_change_region(self, event):
        """Set a new region boundary."""
        if "release" not in event.properties["attr"]:
            return
        old_value = event.properties["old_value"][0]
        new_value = event.properties["value"][0]
        spectrum = event.properties["data"][0]
        bg_bounds = spectrum.background_bounds.copy()
        for lower, upper in zip(bg_bounds[0::2], bg_bounds[1::2]):
            if old_value == lower:
                spectrum.remove_background_bounds(lower, upper)
                spectrum.add_background_bounds(new_value, upper)
            if old_value == upper:
                spectrum.remove_background_bounds(lower, upper)
                spectrum.add_background_bounds(lower, new_value)
        self.bus.fire()

    def on_remove_region(self, *_args):
        """Remove selected region."""
        navbar = self.get_widget("plot_toolbar")
        def remove_region(_x_0, _y_0, x_1, _y_1):
            """Remove region"""
            esel = x_1
            for spectrum in self.state.active_spectra:
                bg_bounds = spectrum.background_bounds.copy()
                for lower, upper in zip(bg_bounds[0::2], bg_bounds[1::2]):
                    if esel >= lower and esel <= upper:
                        spectrum.remove_background_bounds(lower, upper)
                if not any(spectrum.background_bounds):
                    spectrum.background_type = "none"
            self.bus.fire()
        navbar.get_point(remove_region)

    def on_add_peak(self, *_args):
        """Add peak to active regions."""
        navbar = self.get_widget("plot_toolbar")
        shape_combo = self.get_widget("new_peak_model_combo")
        shape = shape_combo.get_active_text()
        def add_peak(position, height, angle):
            """Create new peak from drawn parameters."""
            for spectrum in self.state.active_spectra:
                height -= spectrum.background_of_E(position)
                name = self.state.next_peak_name
                peak = spectrum.add_peak(name, position=position, angle=angle,
                                         height=height, shape=shape)
                peak.register_queue(self.bus)
            self.bus.fire()
        wedgeprops = {}
        navbar.get_wedge(add_peak, **wedgeprops)

    def on_remove_active_peak(self, *_args):
        """Remove active peak."""
        for peak in self.state.active_peaks:
            self.state.peak_names.remove(peak.name)
            peak.spectrum.remove_peak(peak)
        self.bus.fire()

    def on_clear_peaks(self, *_args):
        """Remove all peaks from active spectra."""
        # with EventQueue("combine-all"):
        for spectrum in self.state.selected_spectra:
            for peak in spectrum.peaks:
                self.state.peak_names.remove(peak.name)
                spectrum.remove_peak(peak.name)
        self.bus.fire()

    def on_peak_entry_activate(self, *_args):
        """Change the active peak's parameters."""
        active_peaks = self.state.active_peaks
        if len(active_peaks) != 1:
            return
        peak = active_peaks[0]
        entries = {
            "peak_position_entry": "position",
            "peak_area_entry": "area",
            "peak_fwhm_entry": "fwhm",
            "peak_alpha_entry": "alpha"
        }
        constraints = []
        for widget_name, attr in entries.items():
            entry = self.get_widget(widget_name)
            constraint = self.parse_peak_entry(entry.get_text())
            constraint["param_alias"] = attr
            constraints.append(constraint)
        for constraint in constraints:
            peak.set_constraint(**constraint)
        model_combo = self.get_widget("peak_model_combo")
        peak.shape = peak.shapes[model_combo.get_active()]
        self.bus.fire()

    def on_peak_model_changed(self, *_args):
        """Change the model of the active peak."""
        active_peaks = self.state.active_peaks
        if len(active_peaks) != 1:
            return
        peak = active_peaks[0]
        model_combo = self.get_widget("peak_model_combo")
        shape = model_combo.get_active_text()
        peak.shape = shape
        self.bus.fire()

    def on_peak_name_entry_changed(self, *_args):
        """Change the active peak's name."""
        active_peaks = self.state.active_peaks
        if len(active_peaks) != 1:
            return
        peak = active_peaks[0]
        name_entry = self.get_widget("peak_name_entry")
        name = name_entry.get_text()
        peak.label = name
        self.bus.fire()

    def on_fit(self, *_args):
        """Do the fucking fitting."""
        active_spectra = self.state.active_spectra
        for spectrum in active_spectra:
            spectrum.do_fit()
        self.bus.fire()

    @staticmethod
    def parse_peak_entry(param_string):
        """Parse what is entered into a peak entry field."""
        kwargs = {}
        if "<" in param_string or ">" in param_string:
            try:
                kwargs["min"] = float(param_string.split(">")[1].split()[0])
            except IndexError:
                kwargs["min"] = None
            try:
                kwargs["max"] = float(param_string.split("<")[1].split()[0])
            except IndexError:
                kwargs["max"] = None
        else:
            try:
                kwargs["value"] = float(param_string.strip())
            except ValueError:
                kwargs["expr"] = param_string.strip()
        return kwargs


class ViewC(Operator):
    """Contains methods for user initiated view manipulation."""
    def on_show_selected_spectra(self, *_args):
        """Callback for showing all selected spectra."""
        spectra = self.state.selected_spectra
        self.state.active_spectra = spectra
        self.bus.fire()

    def on_spectrum_view_row_activated(self, treeview, path, _col):
        """Callback for row activation by Enter key or double click."""
        model = treeview.get_model()
        iter_ = model.get_iter(path)
        spectrum = model.get(iter_, 0)[0]
        self.state.active_spectra = [spectrum]
        self.bus.fire()

    def on_spectrum_view_clicked(self, treeview, event):
        """Callback for button-press-event, popups the menu on right click.
        Return value determines if the selection on self persists,
        True if not."""
        if not (event.type == Gdk.EventType.BUTTON_PRESS
                and event.button == Gdk.BUTTON_SECONDARY):
            return False
        _, pathlist = treeview.get_selection().get_selected_rows()
        tvmenu = self.get_widget("spectrum_view_context_menu")
        tvmenu.popup(None, None, None, None, event.button, event.time)
        pathinfo = treeview.get_path_at_pos(int(event.x), int(event.y))
        if pathinfo is None:
            return False
        return pathinfo[0] in pathlist

    def on_spectrum_view_filter_changed(self, *_args):
        """Applies search term from entry.get_text() to the TreeView in column
        combo.get_active_text()."""
        combo = self.get_widget("spectrum_view_search_combo")
        entry = self.get_widget("spectrum_view_search_entry")
        self.state.spectra_tv_filter = (
            combo.get_active_text(),
            entry.get_text()
        )
        self.bus.fire()

    def on_peak_view_row_activated(self, treeview, path, _col):
        """Callback for row activation by Enter key or single click."""
        model = treeview.get_model()
        iter_ = model.get_iter(path)
        peak = model.get(iter_, 0)[0]
        self.state.active_peaks = [peak]
        self.bus.fire()

    def on_show_rsfs(self, *_args):
        """Opens an RSF dialog."""
        dialog = self.get_widget("rsf_dialog")
        source_combo = self.get_widget("rsf_combo")
        element_entry = self.get_widget("rsf_entry")
        source_combo.set_active_id(self.state.photon_source_id)
        element_entry.set_text(" ".join(self.state.rsf_elements))
        response = dialog.run()
        if response == Gtk.ResponseType.APPLY:
            elements = re.findall(r"[\w]+", element_entry.get_text())
            self.state.rsf_elements = [element.title() for element in elements]
            self.state.photon_source = source_combo.get_active_text()
        elif response == Gtk.ResponseType.REJECT:
            self.state.rsf_elements = []
        dialog.hide()
        self.bus.fire()

    def on_center_plot(self, *_args):
        """Centers the plot via the navbar command."""
        navbar = self.get_widget("plot_toolbar")
        navbar.center()

    def on_pan_plot(self, *_args):
        """Activates plot panning."""
        navbar = self.get_widget("plot_toolbar")
        navbar.pan()

    def on_zoom_plot(self, *_args):
        """Activates plot panning."""
        navbar = self.get_widget("plot_toolbar")
        navbar.zoom()



"""Starts the program."""


def main():
    """Runs GXPS app from gxps.gxps module."""
    # Test for some cli options immediately
    import sys
    if "--version" in sys.argv:
        version = "unknown version"
        try:
            from gxps import __version__
            version = __version__
        except ImportError:
            if sys.platform == "win32":
                version = "windows-release"
            else:
                try:
                    from pbr.version import VersionInfo
                    version = VersionInfo("gxps").version_string()
                except ImportError:
                    pass
        print(version)
        sys.exit(0)
    if "--release" in sys.argv:
        release = "unknown release"
        try:
            from gxps import __release__
            release = __release__
        except ImportError:
            if sys.platform == "win32":
                release = "windows-release"
            else:
                try:
                    from pbr.version import VersionInfo
                    version = VersionInfo("gxps").release_string()
                except ImportError:
                    pass
        print(release)
        sys.exit(0)


    # Set process name
    if sys.platform == 'linux':
        # pylint: disable=broad-except
        try:
            import ctypes
            libc = ctypes.CDLL('libc.so.6')
            libc.prctl(15, b'gxps', 0, 0, 0)  # 15 = PR_SET_NAME
        except Exception:
            pass
    try:
        from setproctitle import setproctitle
        setproctitle('gxps')
    except ImportError:
        pass

    # Test dependencies
    try:
        from gi.repository import GLib
    except (ImportError, ModuleNotFoundError):
        print("-" * 79)
        print(
            "There are some missing dependencies, please do:"
            "$ sudo apt install python-gi libgirepository1.0-dev"
            "python3-gi-cairo gir1.2-gtk-3.0"
        )
        print("-" * 79)
        raise

    # Set up config and logging
    from gxps.config import (
        activate_logging,
        load_configs
    )
    load_configs()
    activate_logging()

    # Build app object
    from gxps.gui import GXPS
    app = GXPS()

    # Run
    import signal
    if hasattr(GLib, "unix_signal_add"):
        GLib.unix_signal_add(GLib.PRIORITY_DEFAULT, signal.SIGINT, app.on_quit)
    exit_status = app.run(sys.argv)
    sys.exit(exit_status)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""GXPS launcher."""

from pathlib import Path
import os
import sys

from gxps.main import main

def set_gxps_env():
    """Find the location of gxps's working directory and insert it to sys.path
    """
    # pylint: disable=no-member
    basedir = Path(__file__).resolve().parents[0]
    if not (basedir / "gxps.py").is_file():
        basedir = Path().cwd()
        if not (basedir / "gxps.py").is_file():
            print("WARNING : gxps source path not found")
    sys.path.insert(0, str(basedir))
    os.environ["GXPS_DIR"] = str(basedir)


if __name__ == "__main__":
    set_gxps_env()
    main()



"""Models for the peaks."""
# pylint: disable=invalid-name
# pylint: disable=logging-format-interpolation
# pylint: disable=abstract-method
# pylint: disable=too-many-arguments

import logging

import numpy as np
import scipy.special as ss
from lmfit.model import Model
from lmfit.models import guess_from_peak, update_param_vals


LOG = logging.getLogger(__name__)

s2 = np.sqrt(2)
s2pi = np.sqrt(2 * np.pi)
ln2 = 1 * np.log(2)
sln2 = np.sqrt(ln2)
s2ln2 = np.sqrt(2 * ln2)
sqrtln2 = np.sqrt(ln2)
tiny = 1e-5

# ideas for further shapes: sech2-function


def pure_ds(x, amplitude=1.0, center=0.0, fwhm=1.0, asym=0.5):
    """Roughly taken from
    https://rdrr.io/github/GSperanza/RxpsG/src/R/XPSFitAlgorithms.r
    """
    sigma = max(fwhm / 2, tiny)
    arg = center - x
    am1 = 1 - asym
    y = (
        amplitude / np.pi * ss.gamma(am1)
        / (arg**2 + sigma**2) ** (am1 / 2)
        * np.cos(np.pi * asym / 2 + am1 * np.arctan(arg / sigma))
    )
    return y

def gaussian(x, amplitude=1.0, center=0.0, fwhm=1.0):
    """Standard gaussian with amplitude = area."""
    sigma = max(tiny, fwhm / (2 * s2ln2))
    arg = center - x
    y = amplitude / (s2pi * sigma) * np.exp(-arg**2 / (2 * sigma**2))
    return y

def lorentzian(x, amplitude=1.0, center=0.0, fwhm=1.0):
    """Standard lorentzian with amplitude = area."""
    gamma = max(tiny, fwhm / 2)
    arg = center - x
    y = amplitude / (gamma * np.pi) * gamma**2 / (arg**2 + gamma**2)
    return y

def voigt(x, amplitude=1.0, center=0.0, fwhm=1.0, fwhm_l=None):
    """Voigt function using Faddeeva function wofz.
    https://en.wikipedia.org/wiki/Voigt_profile
    Taken from lmfit module, modified to take fwhms:
    Gaussian fwhm and Lorentzian fwhm_l
    """
    if fwhm_l is None:
        fwhm_l = fwhm
    sigma = max(tiny, fwhm / (2 * s2ln2))
    gamma = max(tiny, fwhm_l / 2)
    arg = center - x
    z = (arg + 1j * gamma) / (sigma * s2)
    # pylint: disable=no-member
    y = amplitude * ss.wofz(z).real / (sigma * s2pi)
    return y

def voigt_defined_fwhm(x, amplitude=1.0, center=0.0, fwhm=1.0, fwhm_g=None):
    """Voigt function using Faddeeva function wofz.
    https://en.wikipedia.org/wiki/Voigt_profile
    Taken from lmfit module, modified to take fwhms:
    Full fwhm and Gaussian fwhm (Lorentzian fwhm is inferred, see bottom
    of wikipedia link)
    """
    if fwhm_g is None:
        fwhm_g = fwhm / 1.6376
    sigma = max(tiny, fwhm_g / (2 * s2ln2))
    fwhm_l = 7.72575 * fwhm - np.sqrt(45.23566 * fwhm**2 + 14.4514 * fwhm_g**2)
    gamma = max(tiny, fwhm_l / 2)
    arg = center - x
    z = (arg + 1j * gamma) / (sigma * s2)
    # pylint: disable=no-member
    y = amplitude * ss.wofz(z).real / (sigma * s2pi)
    return y

def gl_sum(x, amplitude=1.0, center=0.0, fwhm=1.0, fraction=0.5):
    """Sum of a gaussian and a lorentzian component."""
    G = gaussian(x, amplitude=amplitude, center=center, fwhm=fwhm)
    L = lorentzian(x, amplitude=amplitude, center=center, fwhm=fwhm)
    return (1 - fraction) * G + fraction * L

def gl_prod(x, amplitude=1.0, center=0.0, fwhm=1.0, fraction=0.5):
    """Product form of a gaussian and a lorentzian component."""
    # area and fwhm are not determined - don't use!
    fwhm_g = fwhm / (1 - fraction)
    fwhm_l = fwhm / fraction
    sigma = max(tiny, fwhm_g / (2 * s2ln2))
    gamma = max(tiny, fwhm_l / 2)
    arg = center - x
    # pylint: disable=no-member
    norm_area = (
        gamma * np.exp(gamma**2 / (4 * sigma**2))
        * ss.erfc(gamma / (2 * sigma))
        / (s2pi * sigma * gamma)
    )
    amplitude /= norm_area
    GL = (
        amplitude #/ (s2pi * sigma * gamma * np.pi)
        * np.exp(-arg**2 / (2 * sigma**2) * 4 * ln2 * (1 - fraction))
        * gamma**2 / (4 * fraction) / (arg**2 + gamma**2 / (4 * fraction))
    )
    # G = gaussian(x, amplitude=amplitude, center=center, fwhm=fwhm_g)
    # L = lorentzian(x, amplitude=amplitude, center=center, fwhm=fwhm_l)
    return GL

def gelius(x, amplitude=1.0, center=0.0, fwhm=1.0, a=0.5, b=0.5, fwhm_l=0.5):
    """See http://www.casaxps.com/help_manual/line_shapes.htm
    Modified to use Voigt profile instead of GL product"""
    if fwhm_l is None:
        fwhm_l = fwhm
    arg = center - x
    below_c = (x <= center).astype(int)
    AW = np.exp(-(2 * sln2 * arg / (fwhm - a * 2 * sln2 * arg))**2)
    w = b * (0.7 + 0.3 / (a + 0.01))
    V = voigt(x, amplitude=amplitude, center=center, fwhm=fwhm, fwhm_l=fwhm_l)
    G = gaussian(x, amplitude=amplitude, center=center, fwhm=fwhm)
    y = V + below_c * (w * (AW - G))
    return y

def asymm_tail(x, center=0, fwhm=1.0, tail=1.0):
    """Tail for dampening asymmetric lines below x = center."""
    arg = (center - x) / fwhm
    try:
        zeros = np.zeros(len(x))
    except TypeError:
        zeros = 0
    y = np.exp(-np.maximum(arg, zeros) * tail)
    return y

def centered_ds(x, amplitude=1.0, center=0.0, fwhm=1.0, asym=0.5):
    """DS lineshape with maximum at center."""
    emax = fwhm / (2 * np.tan(np.pi / (2 - asym)))
    center += emax
    y = pure_ds(x, amplitude=amplitude, center=center, fwhm=fwhm, asym=asym)
    return y

def tailed_ds(x, amplitude=1.0, center=0.0, fwhm=1.0, asym=0.5, tail=1.0):
    """Centered DS with exponential tail at lower x."""
    emax = fwhm / (2 * np.tan(np.pi / (2 - asym)))
    center += emax
    ds = pure_ds(x, amplitude=amplitude, center=center, fwhm=fwhm, asym=asym)
    as_tail = asymm_tail(x, center=center, fwhm=fwhm, tail=tail)
    return ds * as_tail


class PeakModel(Model):
    """Generic model for peaks."""
    def __init__(self, func, **kwargs):
        kwargs["independent_vars"] = kwargs.get("independent_vars", ["x"])
        kwargs["prefix"] = kwargs.get("prefix", "")
        kwargs["nan_policy"] = kwargs.get("nan_policy", "raise")
        self.fwhm_res = kwargs.get("fwhm_res", 0.01)
        self.area_res = kwargs.get("area_res", 0.1)
        self.area_range = kwargs.get("area_range", 20)
        for arg in ("amplitude", "fwhm", "center"):
            if arg not in func.__code__.co_varnames:
                raise ValueError("Function has wrong parameters for PeakModel")
        super().__init__(func, **kwargs)

    def guess(self, data, **kwargs):
        """Guess the pars."""
        x = kwargs.get("x", None)
        negative = kwargs.get("negative", False)
        pars = guess_from_peak(self, data, x, negative, ampscale=0.5)
        return update_param_vals(pars, self.prefix, **kwargs)

    def get_fwhm(self, params, x=None):
        """Generic FWHM calculator:
        Searches from center in both directions for values below maximum / 2
        """
        if x:
            pass
        funcargs = self.make_funcargs(params)
        center = funcargs["center"]
        fwhm = funcargs["fwhm"]
        hm = self.func(x=center, **funcargs) / 2
        x_min, x_max = center, center
        while self.func(x=x_max, **funcargs) >= hm:
            x_max += self.fwhm_res
            if x_max > center + 5 * fwhm:
                LOG.warning("Could not calculate correct FWHM")
                break
        while self.func(x=x_min, **funcargs) >= hm:
            x_min -= self.fwhm_res
            if x_min < center - 5 * fwhm:
                LOG.warning("Could not calculate correct FWHM")
                break
        return x_max - x_min

    def get_area(self, params, x=None):
        """Generic area calculator: Integrates interval
        (center - self.area_range/2, center + self.area_range/2)
        with resolution self.area_res.
        """
        funcargs = self.make_funcargs(params)
        if x:
            start = x.min()
            end = x.max()
            N = len(x)
            res = (end - start) / N
        else:
            center = funcargs["center"]
            start = center - self.area_range / 2
            end = center + self.area_range / 2
            N = self.area_range / self.area_res
            res = self.area_res
        x = np.linspace(start, end, int(N))
        y = self.func(x=x, **funcargs)
        return sum(y) * res


class VoigtModel(PeakModel):
    """Voigt model with a defined fwhm."""
    def __init__(self, **kwargs):
        super().__init__(voigt_defined_fwhm, **kwargs)

class PseudoVoigtModel(PeakModel):
    """Standard Gaussian-Lorentzian product."""
    def __init__(self, **kwargs):
        super().__init__(gl_sum, **kwargs)

class DoniachSunjicModel(PeakModel):
    """x-axis reversed Doniach model (general formula taken from lmfit)."""
    def __init__(self, **kwargs):
        super().__init__(centered_ds, **kwargs)

class TailedDoniachSunjicModel(PeakModel):
    """DS line shape with an exponentially decaying tail on the asymmetric
    side."""
    def __init__(self, **kwargs):
        super().__init__(tailed_ds, **kwargs)



def pah2fwhm(_position, angle, height, shape):
    """Calculates fwhm from position, angle, height depending on shape."""
    if shape == "PseudoVoigt":
        return np.tan(angle) * height
    elif shape == "DoniachSunjic":
        return np.tan(angle) * height
    elif shape == "Voigt":
        return np.tan(angle) * height
    raise NotImplementedError

def pah2area(_position, angle, height, shape):
    """Calculates area from position, angle, height depending on shape."""
    if shape == "PseudoVoigt":
        fwhm = np.tan(angle) * height
        area = (height * (fwhm * np.sqrt(np.pi / ln2))
                / (1 + np.sqrt(1 / (np.pi * ln2))))
        return area
    elif shape == "DoniachSunjic":
        fwhm = np.tan(angle) * height
        area = height / pure_ds(0, amplitude=1, center=0, fwhm=fwhm, asym=0.5)
        return area
    elif shape == "Voigt":
        fwhm = np.tan(angle) * height
        area = height / voigt(0, amplitude=1, center=0, fwhm=fwhm, fwhm_l=0.5)
        return area
    raise NotImplementedError



"""Provides utility classes for Observer pattern and Singleton pattern."""
# pylint: disable=logging-format-interpolation

import logging
import copy


LOG = logging.getLogger(__name__)


class MetaDataContainer:
    """For inheriting: enables quick access to a metadata dictionary
    that prevents the object attributes from becoming too cluttered while
    at the same time offering a hook for doing something upon
    changing or calling a meta datum.
    """
    def __init__(self):
        self._meta = {}
        self._default_meta_value = "RAISE"
        super().__init__()

    def _get_meta(self, attr):
        """Hook being invoked on every self.get_meta call."""
        pass

    def get_meta(self, attr, silent=False):
        """Gives a datum from the dictionary and executes the self._get_meta
        hook.
        """
        if not silent:
            self._get_meta(attr)
        if attr not in self._meta:
            if self._default_meta_value == "RAISE":
                raise AttributeError(
                    "No meta-attribute '{}' exists in {}"
                    "".format(attr, self)
                )
            return self._default_meta_value
        return self._meta[attr]

    def _set_meta(self, attr, value):
        """Hook being invoked on every self.set_meta call."""
        pass

    def set_meta(self, attr, value, silent=False):
        """Sets a datum in the dictionary and executes the self._set_meta
        hook afterwards.
        """
        if not silent:
            self._set_meta(attr, value)
        self._meta[attr] = value


class Event:
    """Stores variables during emitting.
    """
    # pylint: disable=too-few-public-methods

    def __init__(self, signal):
        self.properties = {}

        self._signal = signal
        self.source = []

    @property
    def signal(self):
        """Expose the signal."""
        return self._signal


class EventList(Event):
    """A list of events that behaves like an event but has all the
    childrens' attribute values.
    """
    # pylint: disable=too-few-public-methods

    def __init__(self, eventlist):
        self._list = eventlist
        super().__init__(self._list[0].signal)
        for event in self._list:
            if event.signal != self._list[0].signal:
                raise AttributeError("EventList has only one signal")
            for prop in event.properties:
                if prop not in self.properties:
                    self.properties[prop] = []
                if isinstance(event.properties[prop], list):
                    self.properties[prop].extend(event.properties[prop])
                else:
                    self.properties[prop].append(event.properties[prop])
            self.source.extend(event.source)


class EventBus:
    """Context manager for queuing Events emitted by Observables so they
    are not emitted multiple times by one method call.
    Structure:
    self._policy[signal] = policy
        if signal not in self._policy, apply default_policy
    self._queue[signal] = [event, ...]
        key "signal" is created when the first event enqueues for it
        when the signal is fired, it's queue is emptied
    self._subscribers[signal] = [(callback, prio), ...]
        upon firing a signal, all appropriate callbacks are called
        ordered by prio
    TODO: don't call the same subscriber multiple times in one "all" firing
    """
    _policies = (
        "ignore",
        "accumulate",
        "fire"
    )

    def __init__(self, default_policy):
        if default_policy not in self._policies:
            raise ValueError("Unknown Event Queue policy")
        self._policy = {"default": default_policy, "all": None}
        self._subscribers = {}
        self._queue = {}
        super().__init__()

    def subscribe(self, callback, signal, priority=5):
        """Connects a callback to the queue.
        """
        if signal == "all":
            raise ValueError("Cannot subscribe to all signals")
        if signal not in self._subscribers:
            self._subscribers[signal] = []
        self._subscribers[signal].append((callback, priority))
        LOG.debug(
            "{} subscribed to signal '{}' on {}"
            "".format(callback, signal, self)
        )

    def unsubscribe(self, callback, signal="all"):
        """Disconnects a callback from the queue, by default for all signals.
        """
        if signal == "all":
            for sub_list in self._subscribers:
                dict(sub_list).pop(callback, None)
                # if callback in sub_list:
                #     sub_list.remove(callback)
            LOG.debug(
                "{} unsubscribed from all signals on {}"
                "".format(callback, self)
            )
        else:
            dict(self._subscribers[signal]).pop(callback, None)
            # if callback in self._subscribers[signal]:
            #     self._subscribers[signal].remove(callback)
            LOG.debug(
                "{} unsubscribed from signal '{}' on {}"
                "".format(callback, signal, self)
            )

    def enqueue(self, event):
        """Enqueues event.
        """
        LOG.debug("Enqueue signal '{}' to {}".format(event.signal, self))
        policy = self._policy.get(event.signal, self._policy["default"])
        if self._policy["all"]:
            policy = self._policy["all"]
        elif policy == "ignore":
            return
        if event.signal not in self._queue:
            self._queue[event.signal] = []
        self._queue[event.signal].append(event)
        if "noqueue" in event.properties or policy == "fire":
            self.fire(event.signal)

    def fire(self, signal="all"):
        """Fires all accumulated events of given signal.
        """
        if signal == "all":
            for specific_signal in self._queue.copy():
                if specific_signal == "all":
                    continue
                self.fire(specific_signal)
        else:
            if signal not in self._queue or not self._queue[signal]:
                return
            if signal not in self._subscribers:
                return
            LOG.debug("Fire signal '{}' on {}".format(signal, self))
            event_list = EventList(self._queue[signal])
            self._queue[signal].clear()
            subs = self._subscribers[signal]
            for callback, _prio in sorted(subs, key=lambda x: x[1]):
                callback(event_list)

    def set_policy(self, policy, signal="all"):
        """Sets a policy for how to act on incoming events.
        """
        if policy not in self._policies:
            raise ValueError("Unknown Event Queue policy")
        self._policy[signal] = policy
        LOG.debug(
            "Set policy to '{}' (signal '{}') on {}"
            "".format(policy, signal, self)
        )

    def get_policy(self, signal="all"):
        """Get specific policy. Useful if it is unknown if the signal exists.
        """
        if signal not in self._policy:
            return self._policy["default"]
        return self._policy[signal]



class Observable:
    """Provides methods for observing these objects via callbacks.
    """
    _signals = ()
    _default_signals = ("registered-observable", )

    def __init__(self, *args, **kwargs):
        self._signals = self._signals + self._default_signals
        self._observers = dict((signal, dict()) for signal in self._signals)
        self._queues = []
        super().__init__(*args, **kwargs)

    @property
    def children(self):
        """To be implemented by inherited classes. All references to
        other Observable objects held inside the class should be given here.
        Has to be strictly hierarchical!
        """
        return []

    def deepcopy(self):
        """Disconnect from all Observer stuff, copy and then reconnect."""
        queues = self.queues
        children_queues = {}
        def unregister_children(obs):
            """Recursively unregister children and following generations.
            """
            for child in obs.children:
                children_queues[child] = child.queues
                child.unregister_all_queues()
                unregister_children(child)
        unregister_children(self)
        self.unregister_all_queues()
        other = copy.deepcopy(self)
        for queue in queues:
            self.register_queue(queue)
        for child, child_queues in children_queues.items():
            for queue in child_queues:
                child.register_queue(queue)
        return other

    @property
    def signals(self):
        """Makes signals accessible.
        """
        signals = list(self._signals).copy()
        return signals

    @property
    def queues(self):
        """Exposes queues.
        """
        return self._queues.copy()

    def register_queue(self, queue):
        """Registers a queue where events are sent to.
        """
        self._queues.append(queue)
        LOG.debug("{} registered to queue {}".format(self, queue))

    def register_children_to_queue(self, queue):
        """Registers the children with the given queue.
        """
        for child in self.children:
            child.register_children_to_queue(queue)
            child.register_queue(queue)

    def unregister_queue(self, queue):
        """Unregisters a queue.
        """
        self._queues.remove(queue)
        LOG.debug("{} unregistered from queue {}".format(self, queue))

    def unregister_all_queues(self):
        """Unregisters all queues.
        """
        self._queues.clear()
        LOG.debug("{} unregistered all from queues".format(self))

    def emit(self, signal, **kwargs):
        """Emit a signal: call all corresponding observers with an event
        object as argument. Also make the propagators propagate.
        """
        if signal not in self._signals:
            raise ValueError("{} cannot emit signal '{}'".format(self, signal))
        event = Event(signal)
        event.source.append(self)
        for key, value in kwargs.items():
            event.properties[key] = value

        for queue in self._queues:
            queue.enqueue(event)



"""Spectrum class represents spectrum data."""
# pylint: disable=too-many-instance-attributes
# pylint: disable=logging-format-interpolation
# pylint: disable=invalid-name

import logging
import re

import numpy as np
from lmfit import Parameters
from lmfit.models import ConstantModel#, PseudoVoigtModel

from gxps.utility import Observable, MetaDataContainer
from gxps.processing import (
    IgnoreUnderflow,
    intensity_at_energy,
    calculate_background,
    make_equidistant,
    make_increasing,
    calculate_normalization_divisor,
)
from gxps.models import(
    pah2fwhm,
    pah2area,
    DoniachSunjicModel,
    VoigtModel,
    PseudoVoigtModel
)


LOG = logging.getLogger(__name__)


class SpectrumContainer(Observable):
    """
    Manages a list of several spectra.
    """
    _signals = ("changed-spectra", )

    def __init__(self, *args, **kwargs):
        self._spectra = []
        super().__init__(*args, **kwargs)
        LOG.info("{} created".format(self))

    @property
    def children(self):
        """Observable children objects."""
        return self.spectra

    @property
    def spectra(self):
        """Returns spectrum objects."""
        return self._spectra.copy()

    def add_spectrum(self, spectrum=None, **specdict):
        """Adds a spectrum."""
        if not spectrum:
            spectrum = ModeledSpectrum(**specdict)
        self._spectra.append(spectrum)

        LOG.info("Added spectrum {} to {}".format(spectrum, self))
        self.emit("changed-spectra")
        return spectrum

    def remove_spectrum(self, spectrum):
        """Removes a spectrum."""
        LOG.info("Removing spectrum {} from {}".format(spectrum, self))
        self._spectra.remove(spectrum)
        self.emit("changed-spectra")

    def clear(self):
        """Clear all spectra from self."""
        self._spectra.clear()
        LOG.info("Cleared container {}".format(self))
        self.emit("changed-spectra")


class Spectrum(Observable, MetaDataContainer):
    """
    Holds data from one single spectrum.
    """
    _signals = ("changed-spectrum", "changed-spectrum-meta")
    _required = ("energy", "intensity", "name", "filename", "notes")
    _bg_types = ("none", "linear", "shirley", "tougaard")
    _norm_types = ("none", "highest", "high_energy", "low_energy", "manual")

    def __init__(self, **kwargs):
        if not all(a in kwargs for a in self._required):
            raise ValueError("Required attribute(s) missing")
        super().__init__()
        self._default_meta_value = ""

        energy = np.array(kwargs.pop("energy"))
        intensity = np.array(kwargs.pop("intensity"))
        if len(energy) != len(intensity) or energy.ndim != 1:
            raise ValueError("energy and intensity array sizes differ")
        energy, intensity = make_increasing(energy, intensity)
        energy, intensity = make_equidistant(energy, intensity)
        self._energy = energy
        self._intensity = intensity

        self._background = np.zeros(self._energy.shape)
        self._background_type = "none"
        self._background_bounds = np.array([])

        self._energy_calibration = 0
        self._normalization_type = "none"
        self._normalization_divisor = 1.0
        self._normalization_energy = None

        for key, value in kwargs.items():
            self.set_meta(key, value, silent=True)

        LOG.info("Spectrum '{}' created ({})".format(self.name, self))

    def _set_meta(self, attr, value):
        """Ensure that setting meta data creates an event."""
        LOG.info("'{}' changes '{}' to '{}'".format(self, attr, value))
        self.emit("changed-spectrum-meta", attr=attr, value=value)

    @property
    def name(self):
        """Expose metadatum "name"."""
        return self.get_meta("name")

    # Energy-related
    @property
    def energy(self):
        """Energy numpy array."""
        return self._energy + self._energy_calibration

    @property
    def kinetic_energy(self):
        """Kinetic energy numpy array."""
        raise NotImplementedError

    @property
    def photon_energy(self):
        """Photon energy numpy array."""
        raise NotImplementedError

    @property
    def energy_calibration(self):
        """Number by which the energy axis is shifted from raw to displayed."""
        return self._energy_calibration

    @energy_calibration.setter
    def energy_calibration(self, value):
        """Only numbers are valid."""
        if abs(value) == np.inf:
            raise ValueError("Invalid energy calibration value 'np.inf'.")
        self._energy_calibration = value
        LOG.info("'{}' changed energy cal to '{}'".format(self, value))
        self.emit("changed-spectrum", attr="energy_calibration")

    # Intensity-related
    @property
    def intensity(self):
        """Intensity numpy array."""
        return self._intensity / self._normalization_divisor

    def intensity_of_E(self, energy):
        """Intensity at energy."""
        return intensity_at_energy(self.energy, self.intensity, energy)

    @property
    def normalization_type(self):
        """Normalization type string."""
        return self._normalization_type

    @normalization_type.setter
    def normalization_type(self, value):
        """Normalization type has to be in self._norm_types."""
        if self._normalization_type == value:
            return
        if value not in self._norm_types:
            raise ValueError("Invalid normalization type '{}'".format(value))
        self._normalization_type = value
        if self._normalization_type != "manual":
            self._normalization_divisor = calculate_normalization_divisor(
                self.normalization_type,
                self.normalization_divisor,
                self.energy,
                self._intensity
            )
        LOG.info("'{}' changed norm type to '{}'".format(self, value))
        self.emit("changed-spectrum", attr="normalization_type")

    @property
    def normalization_divisor(self):
        """Return divisor for intensity normalization."""
        return self._normalization_divisor

    @normalization_divisor.setter
    def normalization_divisor(self, value):
        """Only numbers are valid. Sets normalization_type to manual."""
        if not abs(value) > 0:
            raise ValueError("Invalid normalization divisor '0.0'")
        self._normalization_type = "manual"
        self._normalization_divisor = value
        LOG.info("'{}' changed norm divisor to '{}'".format(self, value))
        self.emit("changed-spectrum", attr="normalization_divisor")

    @property
    def normalization_energy(self):
        """Energy at which the normalization is done."""
        return self._normalization_energy

    @normalization_energy.setter
    def normalization_energy(self, value):
        """Setting this affects the divisor and type directly."""
        self.normalization_divisor = intensity_at_energy(
            self.energy,
            self._intensity,
            value
        )
        self._normalization_energy = value

    # Background-related
    @property
    def background(self):
        """Background numpy array."""
        return self._background / self._normalization_divisor

    def background_of_E(self, energy):
        """Intensity at energy."""
        return intensity_at_energy(self.energy, self.background, energy)

    @property
    def background_type(self):
        """Background type string."""
        return self._background_type

    @background_type.setter
    def background_type(self, value):
        """Only values from self._bg_types are valid."""
        if value not in self._bg_types:
            raise ValueError("Background type {} not valid".format(value))
        if self._background_type == value:
            return
        self._background_type = value
        self._background = calculate_background(
            self.background_type,
            self.background_bounds,
            self.energy,
            self._intensity
        )
        LOG.info("'{}' changed bg type to '{}'".format(self, value))
        self.emit("changed-spectrum", attr="background_type")

    @property
    def background_bounds(self):
        """Even-length numpy array with boundary values of slices where
        background will be subtracted via method defined in
        self.background_type.
        """
        return self._background_bounds + self._energy_calibration

    @background_bounds.setter
    def background_bounds(self, value):
        """Only even-length numeral sequence-types are valid."""
        if len(value) % 2 != 0:
            raise ValueError("Background bounds must be pairwise.")
        if not any(value):
            self._background_bounds = np.array([])
        for bound in value:
            if bound > self.energy.max() or bound < self.energy.min():
                raise ValueError("Background bound out of energy range.")
        self._background_bounds = np.array(sorted(list(value)))
        self._background_bounds -= self._energy_calibration
        self._background = calculate_background(
            self.background_type,
            self.background_bounds,
            self.energy,
            self._intensity
        )
        LOG.info("'{}' changed bg bounds to '{}'".format(self, value))
        self.emit("changed-spectrum", attr="background_bounds")

    def add_background_bounds(self, emin, emax):
        """Adds one pair of background boundaries."""
        if emin > emax:
            emin, emax = emax, emin
        old_bounds = self.background_bounds
        self.background_bounds = np.append(old_bounds, [emin, emax])

    def remove_background_bounds(self, emin, emax):
        """Removes one pair of background boundaries."""
        if emin > emax:
            emin, emax = emax, emin
        old_bounds = self.background_bounds.copy()
        self.background_bounds = np.setdiff1d(old_bounds, [emin, emax])


class ModeledSpectrum(Spectrum):
    """Holds information on the Fit and provides methods for fitting.
    """
    _signals = ("changed-fit", "changed-spectrum", "changed-spectrum-meta")
    _required = ("energy", "intensity", "name", "filename", "notes")
    _bg_types = ("none", "linear", "shirley", "tougaard")
    _norm_types = ("none", "manual", "highest", "high_energy", "low_energy")

    def __init__(self, *args, **kwargs):
        self.params = Parameters()
        self._peaks = []
        super().__init__(*args, **kwargs)

    @property
    def children(self):
        """Observable children objects."""
        return self.peaks

    @property
    def peaks(self):
        """Returns peaks."""
        return self._peaks.copy()

    @property
    def fit(self):
        """Returns fit result on whole energy range."""
        with IgnoreUnderflow():
            fit = self.model.eval(params=self.params, x=self.energy)
        try:
            if len(fit) != len(self.energy):
                LOG.warning("Fit array has different length as energy array.")
        except TypeError:
            fit = np.zeros(self.energy.shape)
        return fit

    def fit_of_E(self, energy):
        """Returns model intensity at given energy."""
        energy = np.array([energy])
        with IgnoreUnderflow():
            fit = self.model.eval(params=self.params, x=energy)
        return fit

    @property
    def model(self):
        """Returns the sum of all peak models."""
        model = ConstantModel(prefix="BASE_")
        model.set_param_hint("c", vary=False, value=0)
        self.params += model.make_params()

        for peak in self._peaks:
            model += peak.model
        return model

    def do_fit(self):
        """Returns the fitted cps values."""
        with IgnoreUnderflow():
            result = self.model.fit(
                self.intensity - self.background,
                self.params,
                x=self.energy
            )
        self.params.update(result.params)
        LOG.info("'{}' fitted".format(self))
        self.emit("changed-fit")

    def add_peak(self, name, **kwargs):
        """
        Add a peak with given parameters. Valid parameters:
        area, fwhm, position and model specific parameters:
            PseudoVoigt: fraction, gausswidth
        """
        if name in [peak.name for peak in self._peaks]:
            raise ValueError("Peak already exists")
        if "fwhm" not in kwargs and "height" in kwargs and "angle" in kwargs:
            kwargs["fwhm"] = pah2fwhm(
                kwargs["position"],
                kwargs["angle"],
                kwargs["height"],
                kwargs["shape"]
                )
        if "area" not in kwargs and "height" in kwargs:
            kwargs["area"] = pah2area(
                kwargs["position"],
                kwargs["angle"],
                kwargs["height"],
                kwargs["shape"]
                )
        kwargs.pop("angle", None)
        kwargs.pop("height", None)
        peak = Peak(name, self, **kwargs)
        self._peaks.append(peak)
        self.emit("changed-fit", attr="peaks")
        return peak

    def add_guessed(self, name):
        """Add a peak while guessing parameters."""
        raise NotImplementedError

    def remove_peak(self, peak):
        """Remove a peak specified by its name."""
        # pars_to_del = [
        #     par for par in self.params
        #     if re.match(r"{}_[a-z]+".format(peak.name), par)
        # ]
        peak.clear_params()
        self._peaks.remove(peak)
        # for par in pars_to_del:
        #     self.params.pop(par)
        self.emit("changed-fit", attr="peaks")


class Peak(Observable):
    """
    Provides read access to peak parameters and provides methods to
    constrain them.
    Whereever possible, parameter "aliases" are used. Independent of
    model, every peak should have:
        area
        fwhm
        position
    and optionally:
        alpha
        beta
    in the aliases, each of these properties are mapped to "real" parameters
    in a way that they all act similar.
    This ensures a consistent API.
    """
    _signals = ("changed-peak", "changed-peak-meta")
    _default_aliases = {
        "alpha": None,
        "beta": None,
        "gamma": None
    }
    _defaults = {
        "alpha": 0.5,
        "beta": 0,
        "gamma": 0
    }
    shapes = ["PseudoVoigt", "DoniachSunjic", "Voigt"]

    def __init__(
            self, name, spectrum,
            area=None, fwhm=None, position=None,
            shape="PseudoVoigt", alpha=None, beta=None, gamma=None
        ):
        # pylint: disable=too-many-arguments
        super().__init__()
        self.name = name
        self.spectrum = spectrum
        self.params = spectrum.params
        self._shape = shape
        self._label = "Peak {}".format(name)
        if None in (area, fwhm, position):
            raise ValueError("Required attribute(s) missing")

        self._model = None
        self.initialize_model(area, fwhm, position, alpha, beta, gamma)
        LOG.info("Peak '{}' created ({})".format(self.name, self))

    def initialize_model(self, area, fwhm, position, alpha, beta, gamma):
        """Initialize the peak model and the parameters."""
        # pylint: disable=too-many-arguments
        self.clear_params()
        self.param_aliases = {
            "area": "amplitude",
            "fwhm": "fwhm",
            "position": "center"
        }
        if self._shape == "PseudoVoigt":
            self.param_aliases["alpha"] = "fraction"
            if alpha is None:
                alpha = 0.5
            self._model = PseudoVoigtModel(prefix="{}_".format(self.name))
            self._model.set_param_hint("fraction", vary=False, value=alpha)
        elif self._shape == "DoniachSunjic":
            self.param_aliases["alpha"] = "asym"
            if alpha is None:
                alpha = 0.1
            self._model = DoniachSunjicModel(prefix="{}_".format(self.name))
            self._model.set_param_hint("asym", vary=False, value=alpha)
        elif self._shape == "Voigt":
            self.param_aliases["alpha"] = "fwhm_l"
            if alpha is None:
                alpha = 0.5
            self._model = VoigtModel(prefix="{}_".format(self.name))
            self._model.set_param_hint("fwhm_l", vary=False, value=alpha)
        else:
            if "" in (beta, gamma, ):
                pass #only for linter
            raise NotImplementedError("Unkown shape '{}'".format(self._shape))

        self.params += self._model.make_params()
        self.get_param("fwhm").set(value=abs(fwhm), min=0, vary=True)
        self.get_param("amplitude").set(value=abs(area), min=0)
        self.get_param("center").set(value=abs(position), min=0)
        # if self._shape == "PseudoVoigt":
        #     self.get_param("sigma").set(expr="{}_fwhm/2".format(self.name))

    def get_param_by_real_name(self, param_name):
        """Shortcut for getting the Parameter object called
        "peak.name_param_name"
        """
        return self.params["{}_{}".format(self.name, param_name)]

    def get_param(self, param_alias):
        """Even shorter cut for getting the Parameter object
        by the param alias.
        """
        aliases = {**self._default_aliases, **self.param_aliases}
        param_name = aliases.get(param_alias, param_alias)
        if param_name is None:
            raise ValueError(
                "{}s model '{}' does not support Parameter '{}'"
                "".format(self, self.shape, param_alias)
            )
        return self.params["{}_{}".format(self.name, param_name)]

    def clear_params(self):
        """Clear this peaks' parameters from the model."""
        pars_to_del = [
            par for par in self.params
            if re.match(r"{}_[a-z]+".format(self.name), par)
        ]
        for par in pars_to_del:
            self.params.pop(par)

    @property
    def model(self):
        """Returns model."""
        return self._model

    @property
    def label(self):
        """A label, for example denoting the core level."""
        return self._label

    @label.setter
    def label(self, value):
        """Emit a signal when changing the label."""
        self._label = value
        self.emit("changed-peak-meta", attr="label", value=value)

    @property
    def intensity(self):
        """Intensity array over whole energy range of parent spectrum."""
        with IgnoreUnderflow():
            intensity = self._model.eval(
                params=self.params,
                x=self.spectrum.energy
            )
        return intensity

    def intensity_of_E(self, energy):
        """Returns model intensity at given energy."""
        with IgnoreUnderflow():
            intensity = self._model.eval(params=self.params, x=energy)
        return intensity

    def set_constraint(
            self, param_alias,
            value=None, vary=None, min=None, max=None, expr=None
        ):
        """Sets a constraint for param. None values will unset the constraint.
        """
        # pylint: disable=too-many-arguments
        # pylint: disable=redefined-builtin
        param = self.get_param(param_alias)
        old_par = self.get_constraints(param_alias)

        for arg in (value, vary, min, max):
            try:
                # fail if x is not in (None, number)
                if arg and abs(arg) > -1:
                    pass
            except TypeError:
                LOG.warning("Invalid constraint value '{}'".format(arg))
                raise TypeError("Invalid constraint value '{}'".format(arg))

        if min is None:
            min = 0
        if max is None:
            max = np.inf
        if vary is None:
            vary = value is None

        param.set(min=min, max=max, vary=vary, value=value, expr="")

        if expr is not None:
            expr = self.relation2expr(expr, param_alias)
            try:
                param.set(expr=expr, min=-np.inf, max=np.inf)
                self.params.valuesdict()
            except (SyntaxError, NameError, TypeError):
                old_par["expr"] = ""
                param.set(**old_par)
                self.emit("changed-peak")
                LOG.warning("Invalid expression '{}'".format(expr))

        LOG.info("Fit parameter set to '{}'".format(param))
        self.emit("changed-peak")

    def get_constraints(self, param_alias):
        """Returns a string containing min/max or expr."""
        try:
            param = self.get_param(param_alias)
        except ValueError:
            constraints = {
                "value": None,
                "min": -np.inf,
                "max": np.inf,
                "vary": None,
                "expr": ""
            }
            return constraints

        relation = ""
        if param.expr is not None:
            relation = self.expr2relation(param.expr)
        constraints = {
            "value": param.value,
            "min": param.min,
            "max": param.max,
            "vary": param.vary,
            "expr": relation
        }
        return constraints

    def expr2relation(self, expr):
        """Translates technical expr string into a human-readable relation.
        """
        def param_repl(matchobj):
            """Replaces 'peakname_param' by 'peakname'"""
            param_key = matchobj.group(0)
            name = param_key.split("_")[0]
            if self in self.spectrum.peaks:
                return name
            return param_key
        regex = r"\b[A-Za-z][A-Za-z0-9]*_[a-z_]+"
        relation = re.sub(regex, param_repl, expr)
        return relation

    def relation2expr(self, relation, param_alias):
        """Translates a human-readable arithmetic relation to an expr string.
        """
        param_name = self.param_aliases[param_alias]
        def name_repl(matchobj):
            """Replaces 'peakname' by 'peakname_param' (searches
            case-insensitive).
            """
            name = matchobj.group(0)
            name = name.upper()
            if name == self.name.upper():
                raise ValueError("Self-reference in peak constraint")
            if name in [peak.name.upper() for peak in self.spectrum.peaks]:
                return "{}_{}".format(name, param_name)
            return name
        regex = r"\b[A-Za-z][A-Za-z0-9]*"
        expr = re.sub(regex, name_repl, relation)
        return expr

    @property
    def shape(self):
        """Returns peak shape."""
        return self._shape

    @shape.setter
    def shape(self, value):
        """Sets the peak shape."""
        if self._shape == value:
            return
        fwhm = self.get_constraints("fwhm")
        area = self.get_constraints("area")
        position = self.get_constraints("position")
        alpha = self.get_constraints("alpha")
        beta = self.get_constraints("beta")
        gamma = self.get_constraints("gamma")
        # only change shape after getting constraints!
        if value in ("PseudoVoigt", "DoniachSunjic", "Voigt"):
            self._shape = value
        else:
            raise NotImplementedError

        self.initialize_model(
            area["value"], fwhm["value"], position["value"],
            alpha["value"], beta["value"], gamma["value"]
        )
        self.emit("changed-peak")

    def get_area(self):
        """Returns measured area under the peak."""
        return self._model.get_area(self.params)

    def get_fwhm(self):
        """Returns measured fwhm of the peak."""
        return self._model.get_fwhm(self.params)

    @property
    def alpha_name(self):
        """Gives the name of the parameter alpha."""
        if self.shape == "PseudoVoigt":
            return "Fraction"
        if self.shape == "DoniachSunjic":
            return "Asym"
        if self.shape == "Voigt":
            return "Lor. FWHM"
        return None

    @property
    def beta_name(self):
        """Gives the name of the parameter beta."""
        return None

    @property
    def gamma_name(self):
        """Gives the name of the parameter gamma."""
        return None



"""Tests the spectrum module."""
# pylint: disable=invalid-name
# pylint: disable=missing-docstring
# pylint: disable=redefined-outer-name
# pylint: disable=protected-access

import pytest
import numpy as np
import matplotlib.pyplot as plt

from gxps.spectrum import Spectrum, SpectrumContainer
from gxps import io


########## Fixtures

parsed_sdicts = [
    io.parse_spectrum_file("tests/fixtures/TiO2-110-f.txt"),
    io.parse_spectrum_file("tests/fixtures/Al-NE-FeCr-S02p6-Fe2p-3.xy"),
    io.parse_spectrum_file("tests/fixtures/21-04_015-XAS_100.xy"),
]

parsed_spectra = [Spectrum(**sdictlist[-1]) for sdictlist in parsed_sdicts]

@pytest.fixture
def tio2f():
    """TiO2-110-f dataset containing 22 individual spectra."""
    specdicts = io.parse_spectrum_file("tests/fixtures/TiO2-110-f.txt")
    return Spectrum(**specdicts[7])

@pytest.fixture
def simple_spectrum():
    return Spectrum(
        energy=[1, 2, 3],
        intensity=[4, 5, 6],
        filename="fixture_nofile",
        name="fixture",
        key="T 2"
    )

parsed_spectra.append(Spectrum(
    energy=[1, 2, 3],
    intensity=[4, 5, 6],
    filename="fixture_nofile",
    name="fixture",
    key="T 1"
))


########## Test functions

def test_spectrum_constructor():
    with pytest.raises(ValueError):
        Spectrum()
    with pytest.raises(ValueError):
        Spectrum(
            energy=[1, 2, 3], intensity=[1, 2], name="n", filename="f"
        )
    with pytest.raises(ValueError):
        Spectrum(
            energy=[1, 2, 3], intensity=[1, 2, 3, 4], name="n", filename="f"
        )
    with pytest.raises(ValueError):
        Spectrum(
            energy=[[1, 2], [3, 4]],
            intensity=[[3, 3], [2, 2]],
            name="n",
            filename="f"
        )

@pytest.mark.parametrize("spectrum", parsed_spectra)
def test_spectrum_plot(spectrum):
    plt.plot(spectrum.energy, spectrum.intensity)
    suffix = (spectrum.meta.filename.split("/")[-1]).split(".")[0]
    plt.savefig("tests/plot_verification/s_{}.png".format(suffix))
    plt.clf()

def test_spectrum_container():
    class ObservingException(Exception):
        pass
    def cb_exception(*_args):
        raise ObservingException

    specdicts = io.parse_spectrum_file("tests/fixtures/TiO2-110-f.txt")
    spectra = SpectrumContainer()
    for specdict in specdicts:
        spectra.add_spectrum(**specdict)
    assert len(specdicts) == len(spectra)
    spectra.connect("changed-spectrum", cb_exception)
    spectra.connect("changed-peak", cb_exception)
    with pytest.raises(ObservingException):
        spectra.spectra[0].normalization_divisor = 3.1
    with pytest.raises(ObservingException):
        spectrum = spectra.spectra[1]
        peak = spectrum.model.add_peak("p1", area=50, fwhm=0.3, position=3)
        peak.set_constraint("fwhm", max=4.2)
    spectra2 = SpectrumContainer()
    assert len(spectra2) == len(specdicts)
    SpectrumContainer.cleanup()

def test_spectrum_observability(tio2f):
    class ObservingException(Exception):
        pass
    def cb_exception(*_args):
        raise ObservingException

    tio2f.connect("changed-spectrum", cb_exception)
    with pytest.raises(ObservingException):
        tio2f.normalization_divisor = 2.3
    tio2f.connect("changed-metadata", cb_exception)
    with pytest.raises(ObservingException):
        tio2f.meta.filename = "test"
    tio2f.model.add_peak("p2", area=200, fwhm=3, position=500)
    tio2f.connect("changed-fit", cb_exception)
    with pytest.raises(ObservingException):
        tio2f.model.add_peak("p1", area=300, fwhm=2, position=340)
    tio2f.connect("changed-peak", cb_exception)
    with pytest.raises(ObservingException):
        tio2f.model["p2"].set_constraint("fwhm", min=4)

def test_spectrum_background_setters(simple_spectrum):
    with pytest.raises(ValueError):
        simple_spectrum.background_type = "foo"
    with pytest.raises(AttributeError):
        simple_spectrum.energy = [10, 11, 12]
    with pytest.raises(AttributeError):
        simple_spectrum.intensity = [10, 11, 12]
    with pytest.raises(ValueError):
        simple_spectrum.background_bounds = [1.1, 1.4, 1.9]
    with pytest.raises(ValueError):
        simple_spectrum.background_bounds = [1.1, 1.3, 1.5, 3.8]
    simple_spectrum.background_type = "linear"
    simple_spectrum.background_bounds = [1.1, 1.3, 1.8, 2.1]
    for bg_type in simple_spectrum._bg_types:
        if bg_type == "tougaard":
            with pytest.raises(NotImplementedError):
                simple_spectrum.background_type = bg_type
        else:
            simple_spectrum.background_type = bg_type

def test_spectrum_energy_calibration(tio2f):
    with pytest.raises(TypeError):
        tio2f.energy_calibration = "one"
    with pytest.raises(ValueError):
        tio2f.energy_calibration = -np.inf
    assert tio2f.energy_calibration == 0.0
    tio2f.background_bounds = [510, 535]
    tio2f.background_type = "shirley"
    e = tio2f.energy.copy()
    i = tio2f.intensity.copy()
    b = tio2f.background.copy()
    bb = tio2f.background_bounds.copy()
    tio2f.energy_calibration = 4.3
    assert np.allclose(tio2f.energy, e + 4.3)
    assert np.allclose(tio2f.intensity, i)
    assert np.allclose(tio2f.background, b)
    assert np.allclose(tio2f.background_bounds, bb + 4.3)

def test_spectrum_normalization(tio2f):
    with pytest.raises(TypeError):
        tio2f.normalization_divisor = [1, 4]
    with pytest.raises(ValueError):
        tio2f.normalization_divisor = 0.0
    with pytest.raises(ValueError):
        tio2f.normalization_type = 5.3
    with pytest.raises(ValueError):
        tio2f.normalization_type = "foo"
    assert tio2f.normalization_type == "none"
    assert tio2f.normalization_divisor == 1.0
    tio2f.background_bounds = [510, 535]
    tio2f.background_type = "shirley"
    e = tio2f.energy.copy()
    i = tio2f.intensity.copy()
    b = tio2f.background.copy()
    tio2f.normalization_divisor = 3.4
    assert tio2f.normalization_type == "manual"
    assert np.allclose(tio2f.energy, e)
    assert np.allclose(tio2f.intensity * 3.4, i)
    assert np.allclose(tio2f.background * 3.4, b)
    tio2f.normalization_divisor = 1.9
    assert np.allclose(tio2f.intensity * 1.9, i)
    assert np.allclose(tio2f.background * 1.9, b)
    tio2f.normalization_type = "none"
    assert tio2f.normalization_divisor == 1.0
    assert np.allclose(tio2f.intensity, i)

def test_peak_addition(tio2f):
    with pytest.raises(ValueError):
        tio2f.model.add_peak(name="p1", fwhm=2.0, area=600)
    with pytest.raises(TypeError):
        tio2f.model.add_peak(name="p1", fwhm=3.1, area=400, position="3br0")
    assert not tio2f.model
    tio2f.model.add_peak(name="p2", fwhm=2.4, area=300, position=515.3)
    with pytest.raises(ValueError):
        tio2f.model.add_peak(name="p2", fwhm=1.0, area=400, position=530)
    assert len(tio2f.model) == 1
    assert tio2f.model["p2"] in tio2f.model.peaks
    tio2f.model.remove_peak(name="p2")
    assert not tio2f.model

def test_peak_manipulation(tio2f):
    tio2f.model.add_peak(name="p1", fwhm=4.3, area=3500, position=530)
    tio2f.model.add_peak(name="p2", fwhm=4.3, area=3500, position=330)
    tio2f.model.add_peak(name="p2axl", fwhm=1.3, area=3500, position=515)
    p1 = tio2f.model["p1"]
    p2 = tio2f.model["p2"]
    assert p1.fwhm == 4.3
    p1.area = 4300
    assert p1.area == 4300
    p1.alpha = 0.6
    assert p1.alpha == 0.6
    plt.plot(tio2f.energy, tio2f.intensity)
    plt.plot(tio2f.energy, p1.get_intensity(tio2f.energy))
    plt.plot(tio2f.energy, tio2f.model["p2"].get_intensity(tio2f.energy))
    plt.plot(tio2f.energy, tio2f.model["p2axl"].get_intensity(tio2f.energy))
    plt.savefig("tests/plot_verification/p_1.png")
    plt.clf()
    with pytest.raises(ValueError):
        p1.set_constraint("fwhm", expr="bla")
    with pytest.raises(ValueError):
        p1.set_constraint("fwhm", expr="p1 =")
    p1.set_constraint("fwhm", max=3.0)
    assert tio2f.model.params["p1_fwhm"].max == 3.0
    assert p1.fwhm == 3.0
    p2.set_constraint("fwhm", min=0.1)
    assert p2.get_constraint("fwhm", "min") == 0.1
    tio2f.model["p2"].fwhm = 4.5
    p1.set_constraint("fwhm", expr="p2 * 3")
    assert p1.fwhm == 13.5
    p1.set_constraint("fwhm", expr="")
    assert p1.get_constraint("fwhm", "vary") is True
    p1.set_constraint("fwhm", expr="p2axl*2")
    assert p1.get_constraint("fwhm", "expr") == "p2axl*2"
    assert p1.get_constraint("fwhm", "vary") is False
    with pytest.raises(ValueError):
        p1.set_constraint("fwhm", expr="p1 * 4")

def test_peak_destruction(tio2f):
    tio2f.model.add_peak(name="p1", fwhm=4.3, area=3500, position=530)
    tio2f.model.add_peak(name="p2", fwhm=4.3, area=3500, position=330)
    tio2f.model.add_peak(name="p2axl", fwhm=1.3, area=3500, position=515)
    assert "p1_fwhm" in tio2f.model.params
    tio2f.model.remove_peak("p1")
    assert "p1_fwhm" not in tio2f.model.params

def test_fitting(tio2f):
    p1 = tio2f.model.add_peak(name="p1", fwhm=2.5, area=5000, position=530)
    p2 = tio2f.model.add_peak(name="p2", fwhm=3, area=3500, position=515)
    p3 = tio2f.model.add_peak(name="p3", fwhm=3, area=1000, position=522)
    tio2f.background_bounds = (509, 535)
    tio2f.background_type = "shirley"
    def plot():
        toplot = (
            tio2f.intensity,
            tio2f.background,
            p1.get_intensity(tio2f.energy) + tio2f.background,
            p2.get_intensity(tio2f.energy) + tio2f.background,
            p3.get_intensity(tio2f.energy) + tio2f.background,
            tio2f.model.get_intensity(tio2f.energy) + tio2f.background
        )
        for i in toplot:
            plt.plot(tio2f.energy, i)
    plot()
    plt.savefig("tests/plot_verification/p_prefit.png")
    plt.clf()
    tio2f.model.fit(tio2f.energy, tio2f.intensity - tio2f.background)
    plot()
    plt.savefig("tests/plot_verification/p_postfit.png")
    plt.clf()
    p1.alpha = 1.5
    tio2f.model.fit(tio2f.energy, tio2f.intensity - tio2f.background)
    assert p1.alpha == 1.0



"""Tests the IO functions."""
# pylint: disable=invalid-name
# pylint: disable=missing-docstring
# pylint: disable=redefined-outer-name

import pytest

from gxps import io

@pytest.mark.parametrize("fname", [
    "tests/fixtures/TiO2-110-f.txt",
    "tests/fixtures/Al-NE-FeCr-S02p6-Fe2p-3.xy",
    "tests/fixtures/contains_layer_zeros.txt",
    "tests/fixtures/21-04_015-XAS_100.xy"
])
def test_parse_spectrum_file(fname):
    specdicts = io.parse_spectrum_file(fname)
    assert specdicts
    for specdict in specdicts:
        assert specdict.keys() == specdicts[0].keys()
        assert "intensity" in specdict
        assert "energy" in specdict

def test_parse_invalid_spectrum_file():
    with pytest.raises(FileNotFoundError):
        io.parse_spectrum_file("tests/fixtures/nonexistent.txt")
    with pytest.raises(ValueError):
        io.parse_spectrum_file("tests/fixtures/faulty.xy")
    with pytest.raises(ValueError):
        io.parse_spectrum_file("tests/fixtures/ag-fitted.xpl")



"""Provides access to configparser instances and logging parameters."""
# pylint: disable=missing-docstring
# pylint: disable=logging-format-interpolation

import sys
import configparser
import logging
import logging.config
import traceback

from gxps.xdg import DATA_DIR, CONF_DIR, LOG_FILE, _make_missing_dirs


CONFIG = configparser.ConfigParser()
def load_cfg():
    # TODO look also in CONF_DIRS and copy to "canonical" location
    default = str(DATA_DIR / "config/config.ini")
    user = str(CONF_DIR / "config.ini")
    CONFIG.read([default, user])
def save_cfg():
    with open(str(CONF_DIR / "config.ini"), "w") as cfg_file:
        CONFIG.write(cfg_file)
CONFIG.load = load_cfg
CONFIG.save = save_cfg

COLORS = configparser.ConfigParser()
def load_colors():
    default = str(DATA_DIR / "config/colors.ini")
    user = str(CONF_DIR / "colors.ini")
    COLORS.read([default, user])
def save_colors():
    with open(str(CONF_DIR / "colors.ini"), "w") as color_file:
        COLORS.write(color_file)
COLORS.load = load_colors
COLORS.save = save_colors

LOG_CFG = {}
# TODO use config file https://docs.python.org/2/library/logging.config.html
# def load_log_cfg():
#     LOG_CFG.update(_logger_conf())
# LOG_CFG.load = load_log_cfg
# LOG_CFG.save = lambda *_args: None

def load_configs():
    _make_missing_dirs()
    CONFIG.load()
    COLORS.load()

def activate_logging():
    _make_missing_dirs()
    LOG_CFG.update(_logger_conf())
    logging.config.dictConfig(LOG_CFG)
    logging.getLogger(__name__)
    logging.getLogger("matplotlib").setLevel(logging.WARNING)
    ex_logger = logging.getLogger("ExceptionLogger")
    def exception_handler(type_, value, tback):
        tbstring = traceback.format_tb(tback)
        tbstring = [line.replace("\n", "") for line in tbstring]
        tbstring = (" " * 10).join(tbstring)
        ex_logger.error(
            "Uncaught {}: {}, traceback: '{}'"
            "".format(type_.__name__, value, tbstring)
        )
        sys.__excepthook__(type_, value, tback)
    sys.excepthook = exception_handler
    file_handler = logging.getLogger("").handlers[1]
    file_handler.doRollover()


def _logger_conf():
    """Logger configuration dictionary."""
    confdict = {
        "version": 1,
        "disable_existing_loggers": True,
        "formatters": {
            "verbose": {
                "format": "{levelname:8s} {asctime} "
                          "{lineno:5d}:{name:20s}{message}",
                "style": "{",
            },
            "brief": {
                "format": "{levelname:8s}: {message}",
                "style": "{",
            }
        },
        "handlers": {
            "console": {
                "class":"logging.StreamHandler",
                "level":"WARNING",
                "formatter": "brief",
                "stream": sys.stderr,
            },
            "file": {
                "class": "logging.handlers.RotatingFileHandler",
                "level": "DEBUG",
                "filename": LOG_FILE,
                "formatter": "verbose",
                "maxBytes": 2000000,        # = 2MB
                "backupCount": 5,
            },
        },
        "loggers": {
            "ExceptionLogger": {
                "handlers": ["file"],
                "level": "ERROR",
                "propagate": False,
            },
        },
        "root": {
            "handlers": ["console", "file"],
            "level":"DEBUG",
        },
    }
    return confdict



"""Implementations of Command class.
Command has method self.get_widget for retrieving Gtk widgets and attributes
self.state (representing GUI state) as well as self.data (representing
spectra).
"""
# pylint: disable=wrong-import-position
# pylint: disable=logging-format-interpolation

import logging
# import os
# import weakref
import functools

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk #, Gio, GLib

# from gxps import CONFIG
# import gxps.io
# from gxps.widgets import GXPSImportDialog


LOG = logging.getLogger(__name__)


class Command:
    """Executes user action: Affects the self.data and self.state objects,
    uses self.get_widget.
    """
    def __init__(self, func=None, undo_func=None, redo_func=None, data=None):
        self.func = func
        if self.func is None:
            raise ValueError("Command missing executor function")
        self.undo_func = undo_func
        self.redo_func = redo_func
        self.state = []
        self.data = []
        self.obj = None

    @property
    def undoable(self):
        """Make testing for undoability available."""
        return self.undo_func is not None

    @property
    def redoable(self):
        """Make testing for redoability available."""
        return self.redo_func is not None

    def __call__(self, *args, **kwargs):
        self.state = self.func(*args, **kwargs)
        return self.state

    def undo(self):
        """Call the undo function."""
        if self.undo_func is None:
            raise AttributeError("Command {} has no undoer".format(self))
        self.undo_func(*self.state)

    def redo(self, *args, **kwargs):
        """Call the redo function. If it is None, call the normal executor."""
        if self.undo_func is None:
            self(*args, **kwargs)
        self.redo_func(*self.state)


# def cmd(func):
#     def wrapper(*args, **kwargs):
#         obj = func.__self__
#         a = func(*args, **kwargs)
#         return a
#     return wrapper



class TC:
    def __init__(self):
        self.last = 0
        self.now = 0

    @Command
    def makegreat(self, value):
        self.last = self.now
        self.now = value
        # return x

    @makegreat.undoer
    def makegreat(self):
        self.now = self.last

    def prln(self):
        print("last: {}".format(self.last))
        print("now: {}".format(self.now))

# x = tclass()
# a = x.makegreat
# a(5)
# a.print()
# a.undo()
# x.print()
# a(8)
# x.print()
# a(10)
# x.print()
# a.undo()
# x.print()


    # def setup(self, *args):
    #     """Provide additional data for execution that is needed."""
    #     self.setup_data = args
    #
    # def execute(self):
    #     """Command execution method to be called from outside."""
    #     self.do_execute()
    #     self.done = True
    #
    # def do_execute(self):
    #     """Does the action. Implemented by child classes."""
    #     raise NotImplementedError
    #
    # def undo(self):
    #     """Reverts data / state objects into the state before self.execute().
    #     """
    #     raise NotImplementedError
    #
    # def redo(self):
    #     """If execute requires a lot of cpu time, this is the faster
    #     way compared to a second self.execute().
    #     """
    #     self.execute()


# class ImportCommand(Command):
#     """Imports data."""
#     # pylint: disable=not-a-mapping
#     def __init__(self, *args, **kwargs):
#         super().__init__(*args, **kwargs)
#         self._added_spectra = weakref.WeakSet([])
#         self._fnames = []
#
#     def do_execute(self):
#         data_dir = os.path.expandvars(CONFIG["IO"]["data-dir"])
#         dialog = ImportDialog(self.get_widget("main_window"), data_dir)
#         response = dialog.run()
#         if response == Gtk.ResponseType.OK:
#             CONFIG["IO"]["data-dir"] = dialog.get_current_folder()
#             specdicts = []
#             for fname in dialog.get_filenames():
#                 specdicts.extend(gxps.io.parse_spectrum_file(fname))
#                 self._fnames.append(fname)
#             for specdict in specdicts:
#                 spectrum = self.data.add_spectrum(**specdict)
#                 self._added_spectra.add(spectrum)
#         dialog.destroy()
#
#     def undo(self):
#         for spectrum in self._added_spectra:
#             self.data.remove(spectrum)
#
#     def redo(self):
#         self._added_spectra.clear()
#         specdicts = []
#         for fname in self._fnames:
#             specdicts.extend(gxps.io.parse_spectrum_file(fname))
#         for specdict in specdicts:
#             spectrum = self.data.add_spectrum(**specdict)
#             self._added_spectra.add(spectrum)



"""Tests the proc functions."""
# pylint: disable=invalid-name
# pylint: disable=missing-docstring
# pylint: disable=redefined-outer-name

import pytest
import numpy as np
import matplotlib.pyplot as plt

from gxps.spectrum import Spectrum
from gxps import io
from gxps import processing as proc


########## Fixtures

parsed_sdicts = [
    io.parse_spectrum_file("tests/fixtures/TiO2-110-f.txt"),
    io.parse_spectrum_file("tests/fixtures/Al-NE-FeCr-S02p6-Fe2p-3.xy"),
    io.parse_spectrum_file("tests/fixtures/21-04_015-XAS_100.xy"),
]

parsed_spectra = [Spectrum(**sdictlist[-1]) for sdictlist in parsed_sdicts]

@pytest.fixture
def tio2f():
    """TiO2-110-f dataset containing 22 individual spectra."""
    specdicts = io.parse_spectrum_file("tests/fixtures/TiO2-110-f.txt")
    return Spectrum(**specdicts[3])

parsed_spectra.append(Spectrum(
    energy=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11],
    intensity=[4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14],
    filename="fixture_fname",
    name="fixture",
    key="T 3"
))


########## Test functions

@pytest.mark.parametrize("spectrum", parsed_spectra)
def test_bg_single_funcs(spectrum):
    i1, i2 = int(len(spectrum.energy) / 3), int(len(spectrum.energy) * 2 / 3)
    e, i = spectrum.energy[i1:i2], spectrum.intensity[i1:i2]
    try:
        bg_s = proc.shirley(e, i)
        assert len(bg_s) == len(e)
    except FloatingPointError:
        pass
    bg_l = proc.linear_bg(e, i)
    assert len(bg_l) == len(e)

@pytest.mark.parametrize("spectrum,method", [
    *[(spectrum, "linear") for spectrum in parsed_spectra],
    *[(spectrum, "shirley") for spectrum in parsed_spectra]
])
def test_calculate_background(spectrum, method):
    min_e = spectrum.energy.min()
    span = spectrum.energy.max() - min_e
    bg = proc.calculate_background(
        method,
        np.array([
            int(min_e + span * 0.1),
            int(min_e + span * 0.5),
            int(min_e + span * 0.6),
            int(min_e + span * 0.8)
        ]),
        spectrum.energy,
        spectrum.intensity
    )
    assert len(bg) == len(spectrum.energy)
    plt.plot(spectrum.energy, spectrum.intensity)
    plt.plot(spectrum.energy, bg)
    suffix = (spectrum.meta.filename.split("/")[-1]).split(".")[0]
    plt.savefig("tests/plot_verification/bg_{}_{}.png".format(method, suffix))
    plt.clf()

@pytest.mark.parametrize("spectrum", parsed_spectra)
def test_calculate_normalization_divisor(spectrum):
    e, i = spectrum.energy, spectrum.intensity
    d = spectrum.normalization_divisor
    divisor = proc.calculate_normalization_divisor("highest", d, e, i)
    assert np.isclose(divisor, i.max())
    divisor = proc.calculate_normalization_divisor("high_energy", d, e, i)
    assert np.isclose(divisor, i[-2], rtol=0.5)
    divisor = proc.calculate_normalization_divisor("low_energy", d, e, i)
    assert np.isclose(divisor, i[2], rtol=0.5)
    divisor = proc.calculate_normalization_divisor("manual", 4, e, i)
    assert np.isclose(divisor, 4)
    divisor = proc.calculate_normalization_divisor("none", d, e, i)
    assert np.isclose(divisor, 1.0)



"""Classes managing single representation parts of the app."""
# pylint: disable=wrong-import-position
# pylint: disable=logging-format-interpolation
# pylint: disable=too-few-public-methods

import logging
import re
from itertools import cycle

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, Pango, GLib

from gxps import __appname__
from gxps.config import COLORS
from gxps.io import get_element_rsfs
from gxps.spectrum import Peak


LOG = logging.getLogger(__name__)


class ViewManager():
    """Helper class for instantiating all the GUI manager classes."""
    # pylint: disable=too-many-instance-attributes
    def __init__(self, app, state, data):
        self.bus = app.bus
        self.data = data
        self.state = state
        self.get_widget = app.get_widget

        self._window = self._instantiate_window()
        self._plot = self._instantiate_plot()
        self._spectra_panel = self._instantiate_spectra_panel()
        self._peaks_panel = self._instantiate_peaks_panel()
        self._edit_dialog = self._instantiate_edit_dialog()

        vlines = self.get_widget("canvas_objects")
        vlines.set_bus(self.bus)

    def _instantiate_window(self):
        """Make a window manager."""
        window = Window(self.get_widget, self.state, self.data)
        signals = ("changed-project", )
        for signal in signals:
            self.bus.subscribe(window.update_titlebar, signal)
        return window

    def _instantiate_plot(self):
        """Make a Plot manager."""
        plot = Plot(self.get_widget, self.state, self.data)
        signals = ("changed-active", "changed-rsf", "changed-spectrum",
                   "changed-fit", "changed-peak")
        for signal in signals:
            self.bus.subscribe(plot.update, signal, priority=10)
        return plot

    def _instantiate_spectra_panel(self):
        """Make a SpectraPanelManager."""
        spectra_panel = SpectraPanel(self.get_widget, self.state, self.data)
        signals = (
            "changed-spectrum",
            "changed-spectrum-meta",
            "changed-spectra",
            "changed-active"
        )
        for signal in signals:
            self.bus.subscribe(spectra_panel.update_data, signal)
        self.bus.subscribe(spectra_panel.update_filter, "changed-tv")
        self.bus.subscribe(spectra_panel.update_controls, "changed-spectrum")
        self.bus.subscribe(spectra_panel.update_controls, "changed-active")
        return spectra_panel

    def _instantiate_peaks_panel(self):
        """Make a PeakPanelManager."""
        peaks_panel = PeakPanel(self.get_widget, self.state, self.data)
        signals = (
            "changed-peak",
            "changed-peak-meta",
            "changed-fit",
            "changed-active"
        )
        for signal in signals:
            self.bus.subscribe(peaks_panel.update_data, signal)
            self.bus.subscribe(peaks_panel.update_controls, signal)
        return peaks_panel

    def _instantiate_edit_dialog(self):
        """Make a EditDialogManager."""
        edit_dialog = EditDialog(self.get_widget, self.state, self.data)
        self.bus.subscribe(edit_dialog.update, "changed-editing-spectra")
        return edit_dialog


class View:
    """Meta class for objects that contain the main functions of the
    application as methods.
    """
    def __init__(self, get_widget, state, data):
        self.get_widget = get_widget
        self.state = state
        self.data = data


class Window(View):
    """Manages all things that regard the behavior and appearance of the
    main window.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        statusbar = self.get_widget("statusbar")
        self._statusbar_id = statusbar.get_context_id("")

    def update_titlebar(self, _event):
        """Updates the title bar to show the current project."""
        fname = self.state.current_project
        isaltered = self.state.project_isaltered
        win = self.get_widget("main_window")
        if fname:
            if isaltered:
                fname += "*"
            win.set_title(u"{} — {}".format(fname, __appname__))
        else:
            win.set_title(__appname__)

    def display(self, event):
        """Updates the statusbar message."""
        message = event.message
        try:
            if event.do_log:
                LOG.info("statusbar: {}".format(message))
        except AttributeError:
            pass
        statusbar = self.get_widget("statusbar")
        message_id = statusbar.push(self._statusbar_id, message)
        def erase_message():
            """Pop message from the statusbar."""
            statusbar.remove(self._statusbar_id, message_id)
            return False
        try:
            GLib.timeout_add_seconds(event.timeout, erase_message)
        except AttributeError:
            pass


class Plot(View):
    """Draws the spectra onto the canvas. Listens to GUIState for what
    exactly to draw.
    """
    # pylint: disable=too-many-instance-attributes
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._canvas = self.get_widget("main_canvas")
        self._ax = self._canvas.ax

    def update(self, event, keepaxes=True):
        """Updates plot by redrawing the whole thing. Relies on
        GUIState for information on what to plot. Designed as a
        callback function for when the plot should change.
        """
        if event.signal == "changed-spectrum":
            if not set(event.source) & set(self.state.active_spectra):
                return
            for attr in event.properties["attr"]:
                if attr in ("normalization_type", "normalization_divisor"):
                    keepaxes = False
        if event.signal == "changed-active":
            keepaxes = False
        # Save axis limits if needed, wipe the canvas and prepare for new
        # centering axis limits.
        if keepaxes:
            self._canvas.store_xylims()
        self._ax.cla()
        self._canvas.reset_xy_centerlims()
        self._plot_spectra()
        self._plot_peaks()
        self._plot_rsf()
        # Either restore axis limits or center plot.
        if keepaxes:
            self._canvas.restore_xylims()
        else:
            self._canvas.center_view()
        self._canvas.draw_idle()
        navbar = self.get_widget("plot_toolbar")
        navbar.disable_tools()

    def _plot_spectra(self):
        colors = cycle(COLORS["Plotting"]["spectra"].split(","))
        vlines = self.get_widget("canvas_objects")
        vlines.clear()
        for spectrum in self.state.active_spectra:
            color = next(colors).strip()
            line = {
                "color": color,
                "linewidth": 1,
                "linestyle": "-",
                "alpha": 1
            }
            self._ax.plot(spectrum.energy, spectrum.intensity, **line)
            line = {
                "color": COLORS["Plotting"]["region-vlines"],
                "linewidth": 2,
                "linestyle": "--",
                "alpha": 1
            }
            for bound in spectrum.background_bounds:
                linewidget = self._ax.axvline(bound, 0, 1, **line)
                vlines.add_line(linewidget, spectrum)
            line = {
                "color": COLORS["Plotting"]["region-background"],
                "linewidth": 1,
                "linestyle": "--"
            }
            if any(spectrum.background):
                self._ax.plot(spectrum.energy, spectrum.background, **line)

            self._canvas.update_xy_centerlims(
                min(spectrum.energy),
                max(spectrum.energy),
                min(spectrum.intensity),
                max(spectrum.intensity)
            )

    def _plot_peaks(self):
        inactive_color = COLORS["Plotting"]["peak"]
        active_color = COLORS["Plotting"]["peak-active"]
        sum_color = COLORS["Plotting"]["peak-sum"]
        for spectrum in self.state.active_spectra:
            for peak in spectrum.peaks:
                line = {}
                if peak in self.state.active_peaks:
                    line = {
                        "color": active_color,
                        "linewidth": 1,
                        "linestyle": "--",
                        "alpha": 0.2
                    }
                    self._ax.fill_between(
                        spectrum.energy,
                        spectrum.background + peak.intensity,
                        spectrum.background,
                        **line
                    )
                else:
                    line = {
                        "color": inactive_color,
                        "linewidth": 1,
                        "linestyle": "--",
                    }
                self._ax.plot(
                    spectrum.energy,
                    spectrum.background + peak.intensity,
                    **line
                )
            line = {
                "color": sum_color,
                "linewidth": 1,
                "linestyle": "--",
            }
            self._ax.plot(
                spectrum.energy,
                spectrum.background + spectrum.fit,
                **line
            )

    def _plot_rsf(self):
        element_colors = cycle(COLORS["Plotting"]["rsf-vlines"].split(","))
        max_rsf = 1e-9
        orbitals = []
        for element in self.state.rsf_elements:
            color = next(element_colors).strip()
            element = get_element_rsfs(element, self.state.photon_source)
            for orbital in element:
                max_rsf = max(orbital["RSF"], max_rsf)
                orbital["color"] = color
            orbitals.extend(element)
        for orbital in orbitals:
            if orbital["RSF"] == 0:
                orbital["RSF"] = max_rsf * 0.5
            self._ax.axvline(
                orbital["BE"],
                0,
                orbital["RSF"] / max_rsf * 0.8,
                color=orbital["color"],
                lw=2
            )
            self._ax.annotate(
                "{} {}".format(orbital["Element"], orbital["Orbital"]),
                xy=(orbital["BE"], orbital["RSF"] / max_rsf * 0.8 + 0.08),
                color=COLORS["Plotting"]["rsf-annotation"],
                xycoords=("data", "figure fraction"),
                ha="center",
                va="bottom"
            )


class SpectraPanel(View):
    """Manages representation of the spectra inside a treeview. Data
    is pulled from the SpectrumContainer and user changeable representation
    attributes are fetched from GUIState.
    Also manages the widgets for manipulating spectra in the same panel.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        tvmenu = self.get_widget("spectrum_view_context_menu")
        treeview = self.get_widget("spectrum_view")
        tvmenu.attach_to_widget(treeview, None)
        norm_combo = self.get_widget("normalization_combo")
        renderer = norm_combo.get_cells()[0]
        renderer.set_property("width-chars", 10)
        self._make_columns()
        self._setup_filter()

    def update_filter(self, event):
        """Updates the filtering with filter parameters from GUIState.
        To be used as callback function.
        """
        if event.signal == "changed-tv":
            if "filter" not in event.properties["attr"]:
                return
        treemodelfilter = self.get_widget("spectrum_filter_treestore")
        treemodelfilter.refilter()

    def update_data(self, event):
        """Updates TreeModel with data from the SpectrumContainer.
        To be used as callback function.
        """
        if event.signal == "changed-active":
            if "spectra" not in event.properties["attr"]:
                return
        selected_spectra = self.state.selected_spectra
        # update model
        treestore = self.get_widget("spectrum_treestore")
        treestore.clear()
        for spectrum in self.data.spectra:
            row = [
                str(spectrum.get_meta(attr))
                for attr in list(self.state.titles["spectrum_view"].keys())
            ]
            treestore.append(parent=None, row=[spectrum] + row)
        # reset selected spectra
        self._set_selection(selected_spectra)

    def update_controls(self, event):
        """Updates the widgets for spectrum manipulation.
        Be careful: Changing a spinbutton value does not equate changing
        the adjustment value. Pay attention to what signals should be emitted,
        for example the updating of the value representation should not
        update the value itself again.
        """
        # pylint: disable=too-many-locals
        # pylint: disable=too-many-branches
        # pylint: disable=too-many-statements
        if event.signal == "changed-active":
            if "spectra" not in event.properties["attr"]:
                return
        active_spectra = self.state.active_spectra
        # background not really part of this panel, but fits best here
        bg_combo = self.get_widget("region_background_type_combo")
        bg_caution = self.get_widget("bg_caution_image")
        cal_spinbutton = self.get_widget("calibration_spinbutton")
        cal_caution = self.get_widget("cal_caution_image")
        norm_combo = self.get_widget("normalization_combo")
        norm_entry = self.get_widget("normalization_entry")
        norm_caution = self.get_widget("norm_caution_image")
        if not active_spectra:
            norm_combo.set_active(-1)
            norm_caution.set_visible(False)
            norm_entry.set_text("")
            norm_entry.set_sensitive(False)
            cal_spinbutton.set_value(0.0)
            cal_caution.set_visible(False)
            bg_combo.set_active(-1)
            bg_caution.set_visible(False)
        else:
            norm_types = [s.normalization_type for s in active_spectra]
            if len(set(norm_types)) > 1:
                norm_combo.set_active(-1)
                norm_caution.set_visible(True)
                norm_entry.set_sensitive(False)
            else:
                normid = self.state.titles["norm_type_ids"][norm_types[0]]
                norm_combo.set_active(int(normid))
                norm_caution.set_visible(False)
                if norm_types[0] == "manual":
                    norm_entry.set_sensitive(True)
                else:
                    norm_entry.set_sensitive(False)
            norm_divs = [s.normalization_divisor for s in active_spectra]
            if len(set(norm_divs)) > 1:
                norm_entry.set_text("")
            else:
                norm_entry.set_text("{:.5f}".format(1 / norm_divs[0]))
            cals = [s.energy_calibration for s in active_spectra]
            if len(set(cals)) > 1:
                cal_spinbutton.set_text("")
                cal_caution.set_visible(True)
            else:
                cal_spinbutton.set_value(cals[0])
                cal_caution.set_visible(False)
            bg_types = [s.background_type for s in active_spectra]
            if len(set(bg_types)) > 1:
                bg_combo.set_active(-1)
                bg_caution.set_visible(True)
            else:
                bgid = self.state.titles["background_type_ids"][bg_types[0]]
                bg_combo.set_active(int(bgid))
                bg_caution.set_visible(False)

    def _set_selection(self, spectra):
        """Selects rows representing spectra."""
        selection = self.get_widget("spectrum_selection")
        treemodelsort = self.get_widget("spectrum_sort_treestore")
        selection.unselect_all()
        for row in treemodelsort:
            if row[0] in spectra:
                selection.select_iter(row.iter)

    def _make_columns(self):
        treeview = self.get_widget("spectrum_view")
        highlight_bg = COLORS["Treeview"]["tv-highlight-bg"]
        lowlight_bg = treeview.style_get_property("even-row-color")
        def render_isplotted(_col, renderer, model, iter_, *_data):
            """Renders the cell light blue if this spectrum is plotted."""
            # render function for making plotted spectra bold and light blue
            spectrum = model.get_value(iter_, 0)
            if spectrum in self.state.active_spectra:
                renderer.set_property("cell-background", highlight_bg)
                renderer.set_property("weight", Pango.Weight.BOLD)
            else:
                renderer.set_property("cell-background", lowlight_bg)
                renderer.set_property("weight", Pango.Weight.NORMAL)
        # the other columns are simple, just apply the render_isplotted func
        for attr in self.state.spectra_tv_columns:
            renderer = Gtk.CellRendererText(xalign=0)
            title = self.state.titles["spectrum_view"][attr]
            # skip first column, it is "spectrum"
            idx = list(self.state.titles["spectrum_view"].keys()).index(attr)
            column = Gtk.TreeViewColumn(title, renderer, text=idx + 1)
            column.set_cell_data_func(renderer, render_isplotted)
            column.set_sort_column_id(idx + 1)
            column.set_resizable(True)
            column.set_reorderable(True)
            treeview.append_column(column)

    def _setup_filter(self):
        # filling the combobox that determines self.tv_filter[0]
        filtercombo = self.get_widget("spectrum_view_search_combo")
        treemodelfilter = self.get_widget("spectrum_filter_treestore")
        for i, meta_attr in enumerate(self.state.spectra_tv_columns):
            title = self.state.titles["spectrum_view"][meta_attr]
            filtercombo.append_text(title)
            if meta_attr == self.state.spectra_tv_filter[0]:
                filtercombo.set_active(i)
        # this function looks into self.tv_filter and executes the regex
        # matching, returning True if the row should be visible
        def filter_func(treemodel, iter_, *_data):
            """Returns True only for rows whose values for the attr
            from self.tv_filter matches the regex from self.tv_filter."""
            meta_attr, search_term = self.state.spectra_tv_filter
            if not meta_attr or not search_term:
                return True
            regex = re.compile(r".*{}.*".format(search_term), re.IGNORECASE)
            # skip first column, it is "spectrum"
            col_index = self.state.spectra_tv_columns.index(meta_attr) + 1
            return re.match(regex, treemodel.get(iter_, col_index)[0])
        treemodelfilter.set_visible_func(filter_func)


class PeakPanel(View):
    """Manages representation of the peaks inside a treeview. Data
    is pulled from the SpectrumContainer and user changeable representation
    attributes are fetched from GUIState.
    TODO: Manage the peak manipulation, e.g. manage constraints
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        shape_combo = self.get_widget("new_peak_model_combo")
        shape_combo.remove_all()
        model_combo = self.get_widget("peak_model_combo")
        model_combo.remove_all()
        for shape in Peak.shapes:
            shape_combo.append_text(shape)
            model_combo.append_text(shape)
        shape_combo.set_active(0)
        self._make_columns()

    def update_data(self, _event):
        """Updates TreeModel with data from the SpectrumContainer.
        To be used as callback function.
        """
        active_peaks = self.state.active_peaks
        # update model
        treestore = self.get_widget("peak_treestore")
        treestore.clear()
        for spectrum in self.state.active_spectra:
            for peak in spectrum.peaks:
                row = [
                    self.format_peak_attr(peak, attr)
                    for attr in list(self.state.titles["peak_view"].keys())
                ]
                treestore.append(parent=None, row=[peak] + row)
        # reset selected spectra
        self._set_selection(active_peaks)

    def update_controls(self, event):
        """Updates the entries representing peak parameters."""
        # pylint: disable=too-many-branches
        active_peaks = self.state.active_peaks
        if event.signal in ("changed-peak", "changed-peak-meta"):
            if not set(event.source) & set(active_peaks):
                return
        if event.signal == "changed-fit":
            if not set(event.source) & set(self.state.active_spectra):
                return

        model_combo = self.get_widget("peak_model_combo")
        entries = {
            "label": self.get_widget("peak_name_entry"),
            "position": self.get_widget("peak_position_entry"),
            "area": self.get_widget("peak_area_entry"),
            "fwhm": self.get_widget("peak_fwhm_entry"),
            "alpha": self.get_widget("peak_alpha_entry"),
            "beta": self.get_widget("peak_beta_entry"),
            "gamma": self.get_widget("peak_gamma_entry"),
            "real_area": self.get_widget("peak_realarea_entry"),
            "real_fwhm": self.get_widget("peak_realfwhm_entry")
        }
        labels = {
            "alpha": self.get_widget("peak_alpha_label"),
            "beta": self.get_widget("peak_beta_label"),
            "gamma": self.get_widget("peak_gamma_label")
        }

        if len(active_peaks) != 1:
            model_combo.set_active(-1)
            model_combo.set_sensitive(False)
            for entry in entries.values():
                entry.set_text("")
                entry.set_sensitive(False)
            for label in labels.values():
                label.set_text("")
            return

        peak = active_peaks[0]
        names = {
            "alpha": peak.alpha_name,
            "beta": peak.beta_name,
            "gamma": peak.gamma_name
        }

        entries["real_area"].set_text("{:.2f}".format(peak.get_area()))
        entries["real_fwhm"].set_text("{:.2f}".format(peak.get_fwhm()))
        entries["label"].set_sensitive(True)
        entries["label"].set_text(peak.label)
        for par in ("position", "area", "fwhm"):
            entries[par].set_sensitive(True)
            entries[par].set_text(self.format_peak_constraints(peak, par))
        for par in ("alpha", "beta", "gamma"):
            if names[par]:
                entries[par].set_sensitive(True)
                entries[par].set_text(self.format_peak_constraints(peak, par))
                labels[par].set_text(names[par])
            else:
                entries[par].set_sensitive(False)
                labels[par].set_text("")

        for i, shape in enumerate(Peak.shapes):
            if shape == peak.shape:
                model_combo.set_active(i)
        model_combo.set_sensitive(True)


    def format_peak_attr(self, peak, attr):
        """Returns a string representing the peak's distinct attribute."""
        if attr in ("position", "fwhm", "area", "alpha", "beta", "gamma"):
            constraints = peak.get_constraints(attr)
            if constraints["value"] is None:
                return ""
            cstring = ""
            if constraints["expr"]:
                cstring += " = {}".format(constraints["expr"])
            elif not constraints["vary"]:
                cstring += " fixed"
            else:
                if constraints["min"] not in (float("-inf"), 0):
                    cstring += " &gt; {} ".format(constraints["min"])
                if constraints["max"] != (float("inf")):
                    cstring += " &lt; {}".format(constraints["max"])
            formatted = (
                "{:.2f}<span color='#999999' font_size='xx-small'> {}</span>"
                "".format(constraints["value"], cstring))
        if attr == "name":
            if len(self.state.active_spectra) >= 2:
                formatted = (
                    "{}<span color='#aaaaaa' font_size='x-small'> ({})</span>"
                    "".format(peak.name, peak.spectrum.name))
            else:
                formatted = str(peak.name)
        if attr in ("label", "shape"):
            formatted = getattr(peak, attr)
        return formatted

    @staticmethod
    def format_peak_constraints(peak, attr):
        """Returns a string representing the peak's attribute's constraints.
        """
        constraints = peak.get_constraints(attr)
        if constraints["value"] is None:
            return ""
        cstring = ""
        if constraints["expr"]:
            cstring += str(constraints["expr"])
        elif not constraints["vary"]:
            cstring += "{}".format(constraints["value"])
        else:
            if constraints["min"] not in (float("-inf"), 0):
                cstring += "> {} ".format(constraints["min"])
            if constraints["max"] != (float("inf")):
                cstring += "< {}".format(constraints["max"])
        return cstring

    def _set_selection(self, peaks):
        """Selects rows representing spectra."""
        selection = self.get_widget("peak_selection")
        treemodelsort = self.get_widget("peak_sort_treestore")
        selection.unselect_all()
        for row in treemodelsort:
            if row[0] in peaks:
                selection.select_iter(row.iter)

    def _make_columns(self):
        treeview = self.get_widget("peak_view")
        highlight_bg = COLORS["Treeview"]["tv-highlight-bg"]
        lowlight_bg = treeview.style_get_property("even-row-color")
        def render_constraints(_col, renderer, model, iter_, idx):
            """Renders the cell light blue if this peak is active."""
            # render function for making plotted spectra bold and light blue
            peak = model.get_value(iter_, 0)
            value = model.get_value(iter_, idx + 1)
            if peak in self.state.active_peaks:
                renderer.set_property("cell-background", highlight_bg)
                renderer.set_property("weight", Pango.Weight.BOLD)
            else:
                renderer.set_property("cell-background", lowlight_bg)
                renderer.set_property("weight", Pango.Weight.NORMAL)
            renderer.set_property("markup", value)
        # the other columns are simple, just apply the render_isplotted func
        for attr in self.state.peak_tv_columns:
            renderer = Gtk.CellRendererText(xalign=0)
            title = self.state.titles["peak_view"][attr]
            idx = list(self.state.titles["peak_view"].keys()).index(attr)
            # skip first column, it is "peak"
            column = Gtk.TreeViewColumn(title, renderer, text=idx + 1)
            column.set_cell_data_func(renderer, render_constraints, idx)
            column.set_sort_column_id(idx + 1)
            column.set_resizable(True)
            column.set_reorderable(True)
            treeview.append_column(column)


class EditDialog(View):
    """Manages the dialog for editing spectra."""
    _exclusion_key = " (mult.)"
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._dialog = self.get_widget("edit_spectrum_dialog")
        self._dialog.exclusion_key = self._exclusion_key

    def update(self, *_args):
        """Updates the dialog to represent the spectra to be edited."""
        spectra = self.state.editing_spectra
        self._dialog.flush()
        def get_value_string(attr, separator=" | "):
            """Returns string to go inside the value fields."""
            if not spectra:
                return ""
            if len(spectra) == 1:
                return str(spectra[0].get_meta(attr))
            values = [str(spectrum.get_meta(attr)) for spectrum in spectra]
            valueset = set(values)
            if len(valueset) > 1:
                valuestring = separator.join(valueset) + self._exclusion_key
            else:
                valuestring = values[0]
            return valuestring

        for attr, title in self.state.titles["static_specinfo"].items():
            try:
                valstring = get_value_string(attr, separator="\n")
            except AttributeError:
                valstring = "NO VALUE SET"
            self._dialog.add_non_editable_row(title, valstring)
        for attr, title in self.state.titles["editing_dialog"].items():
            try:
                valstring = get_value_string(attr, separator="\n")
            except AttributeError:
                valstring = "NO VALUE SET"
            self._dialog.add_editor_row(attr, title, valstring)
